@@ -44,17 +44,21 @@
 #endif /* HAVE_LOCALE_T_IN_XLOCALE_H */
 
 #include "api-types.h"
+#include "dict-common/dialect.h"
 #include "tracon-set.h"
-#include "corpus/corpus.h"
 #include "memory-pool.h"
 #include "string-set.h"
 
 /* Performance tuning.
- * For short sentences, setting suffix IDs and packing takes more
- * resources than it saves. If this overhead is improved, these
- * limit can be set lower. */
+ * For short sentences, encoding takes more resources than it saves. If
+ * this overhead is improved, this limit can be set lower. */
 #define SENTENCE_MIN_LENGTH_TRAILING_HASH 6
 
+/* Pruning per null-count is costly for sentences whose parsing time
+ * is relatively small. If a better pruning per null-count is implemented,
+ * this limit can be set lower. */
+#define SENTENCE_MIN_LENGTH_MULTI_PRUNING 30
+
 typedef struct Cost_Model_s Cost_Model;
 struct Cost_Model_s
 {
@@ -96,7 +100,6 @@ struct Parse_Options_s
 	bool islands_ok;       /* If TRUE, then linkages with islands
 	                          (separate component of the link graph)
 	                          will be generated (default=FALSE) */
-	bool use_cluster_disjuncts; /* Attempt using a broader list of disjuncts */
 	size_t short_length;   /* Links that are limited in length can be
 	                          no longer than this.  Default = 16 */
 	bool all_short;        /* If true, there can be no connectors that are exempt */
@@ -110,6 +113,9 @@ struct Parse_Options_s
 	/* Options governing the generation of linkages. */
 	size_t linkage_limit;  /* The maximum number of linkages processed 100 */
 	bool display_morphology;/* If true, print morpho analysis of words TRUE */
+
+	/* Options governing the dictionary interpretation. */
+	dialect_info dialect;
 };
 
 typedef struct word_queue_s word_queue_t;
@@ -126,17 +132,18 @@ struct Sentence_s
 	size_t length;              /* Number of words */
 	Word  *word;                /* Array of words after tokenization */
 	String_set *   string_set;  /* Used for assorted strings */
-	Pool_desc * fm_Match_node;  /* Fast-matcher Match_node memory pool */
+	Pool_desc * Match_node_pool;
 	Pool_desc * Table_connector_pool; /* Count memoizing memory pool */
-	Pool_desc * E_list_pool;
+	Pool_desc * wordvec_pool;   /* For tracon-word zero-count memoizing */
 	Pool_desc * Exp_pool;
 	Pool_desc * X_node_pool;
 	Pool_desc * Disjunct_pool;
 	Pool_desc * Connector_pool;
 
 	/* Connector encoding, packing & sharing. */
-	size_t min_len_encoding;     /* Do it from this sentence length. */
+	size_t min_len_encoding;     /* Encode from this sentence length. */
 	void *dc_memblock;           /* For packed disjuncts & connectors. */
+	unsigned int num_disjuncts;  /* Number of disjuncts in dc_memblock. */
 
 	/* Wordgraph stuff. FIXME: create stand-alone struct for these. */
 	Gword *wordgraph;            /* Tokenization wordgraph */
@@ -146,6 +153,8 @@ struct Sentence_s
 	size_t gword_node_num;       /* Debug - for differentiating between
 	                                wordgraph nodes with identical subwords. */
 
+	size_t min_len_multi_pruning; /* Do it from this sentence length. */
+
 	/* Parse results */
 	int    num_linkages_found;  /* Total number before postprocessing.  This
 	                               is returned by the do_count() function */
@@ -158,7 +167,7 @@ struct Sentence_s
 	                               This is not the same as num alloced
 	                               because some may be non-canonical. */
 	size_t num_valid_linkages;  /* Number with no pp violations */
-	size_t null_count;          /* Number of null links in linkages */
+	unsigned int null_count;    /* Number of null links in linkages */
 	Linkage        lnkages;     /* Sorted array of valid & invalid linkages */
 	Postprocessor * postprocessor;
 	Postprocessor * constituent_pp;
@@ -169,6 +178,14 @@ struct Sentence_s
 #ifdef USE_SAT_SOLVER
 	void *hook;                 /* Hook for the SAT solver */
 #endif /* USE_SAT_SOLVER */
+
+	/* Generation mode - finding unused dictionary disjuncts.
+	 * This stuff is here and not in Dictionary because it depends on the
+	 * current dialect and cost. */
+	bool *disjunct_used;        /* Used dict disjuncts in memblock below. */
+	void *wildcard_word_dc_memblock;  /* Dictionary disjuncts & connectors. */
+	unsigned int wildcard_word_dc_memblock_sz;
+	unsigned int wildcard_word_num_disjuncts;
 };
 
 #endif
