@@ -14,134 +14,140 @@
  * Miscellaneous utilities for dealing with word types.
  */
 
-#include <math.h>  // for fabs()
-
+#include "api-structures.h"             // Parse_Options
 #include "connectors.h"
 #include "dict-api.h"
 #include "string-set.h"
+#include "dict-defines.h" // for SUBSCRIPT_MARK
 #include "dict-utils.h"
 
+
+/** Replace the right-most dot with SUBSCRIPT_MARK */
+void patch_subscript(char * s)
+{
+	char *ds, *de;
+	int dp;
+	ds = strrchr(s, SUBSCRIPT_DOT);
+	if (!ds) return;
+
+	/* A dot at the end or a dot followed by a number is NOT
+	 * considered a subscript */
+	de = ds + 1;
+	if (*de == '\0') return;
+	dp = (int) *de;
+
+	/* If it's followed by a UTF8 char, its NOT a subscript */
+	if (127 < dp || dp < 0) return;
+	/* assert ((0 < dp) && (dp <= 127), "Bad dictionary entry!"); */
+	if (isdigit(dp)) return;
+	*ds = SUBSCRIPT_MARK;
+}
+
 /* ======================================================== */
 /* Public API ... */
 
 const char * lg_exp_get_string(const Exp* exp)
 {
-	return exp->u.condesc->string;
+	return exp->condesc->string;
 }
 
 /* ======================================================== */
-/* Exp utilities ... */
 
-void free_E_list(E_list *);
-void free_Exp(Exp * e)
+#if 0
+void free_Exp(Exp *e)
 {
-	// Exp might be null if the user has a bad dict. e.g. badly formed
-	// SQL dict.
-	if (NULL == e) return;
-	if (e->type != CONNECTOR_type) {
-		free_E_list(e->u.l);
+	if (NULL == e) return; /* Exp might be null if the user has a bad dict. */
+	Exp *operand_next;
+
+	if (e->type != CONNECTOR_type)
+	{
+		for (Exp *opd = e->operand_first; opd != NULL; opd = operand_next)
+		{
+			operand_next = opd->operand_next;
+			free_Exp(opd);
+		}
 	}
 	free(e);
 }
+#endif
 
-void free_E_list(E_list * l)
-{
-	if (l == NULL) return;
-	free_E_list(l->next);
-	free_Exp(l->e);
-	free(l);
-}
+/* Exp utilities ... */
 
 /* Returns the number of connectors in the expression e */
 int size_of_expression(Exp * e)
 {
-	int size;
-	E_list * l;
+	int size = 0;
+
 	if (e->type == CONNECTOR_type) return 1;
-	size = 0;
-	for (l=e->u.l; l!=NULL; l=l->next) {
-		size += size_of_expression(l->e);
-	}
+	for (Exp *opd = e->operand_first; opd != NULL; opd = opd->operand_next)
+		size += size_of_expression(opd);
+
 	return size;
 }
 
-/**
- * Build a copy of the given expression (don't copy strings, of course)
- */
-static E_list *copy_E_list(E_list *l, Pool_desc* mp[])
-{
-
-	E_list el_head;
-	E_list *el = &el_head;
-
-	for (; l != NULL; l = l->next)
-	{
-		E_list *nl = pool_alloc(mp[0]);
-
-		nl->e = pool_alloc(mp[1]);
-		*nl->e = *l->e;
-		el->next = nl;
-		el = nl;
-		if (l->e->type != CONNECTOR_type)
-			nl->e->u.l = copy_E_list(l->e->u.l, mp);
-	}
-	el->next = NULL;
-
-	return el_head.next;
-}
-
-Exp *copy_Exp(Exp *e, Pool_desc *E_list_pool, Pool_desc *Exp_pool)
+Exp *copy_Exp(Exp *e, Pool_desc *Exp_pool, Parse_Options opts)
 {
 	if (e == NULL) return NULL;
-	Exp *ne = pool_alloc(Exp_pool);
+	Exp *new_e = pool_alloc(Exp_pool);
 
-	*ne = *e;
-	if (CONNECTOR_type == e->type) return ne;
+	*new_e = *e;
+	if ((e->type != CONNECTOR_type) && (Exptag_dialect == e->tag_type))
+		new_e->cost += opts->dialect.cost_table[new_e->tag_id];
 
-	ne->u.l = copy_E_list(ne->u.l, (Pool_desc*[]){E_list_pool, Exp_pool});
-	return ne;
+#if 0 /* Not used - left here for documentation. */
+	new_e->operand_next = copy_Exp(e->operand_next, Exp_pool);
+	if (CONNECTOR_type == e->type) return new_e;
+	new_e->operand_first = copy_Exp(e->operand_first, Exp_pool);
+#else
+	if (CONNECTOR_type == e->type) return new_e;
+
+	/* Iterate operands to avoid a deep recursion due to a lot of operands. */
+	Exp **tmp_e_a = &new_e->operand_first;
+	for(Exp *opd = e->operand_first; opd != NULL; opd = opd->operand_next)
+	{
+		*tmp_e_a = copy_Exp(opd, Exp_pool, opts);
+		tmp_e_a = &(*tmp_e_a)->operand_next;
+	}
+	*tmp_e_a = NULL;
+#endif
+
+	return new_e;
 }
 
 /**
- * Compare two expressions, return true for equal, false for unequal
+ * Compare two expressions, return true for equal, false for unequal.
  */
-static bool exp_compare(Exp * e1, Exp * e2)
+static bool exp_compare(Exp *e1, Exp *e2)
 {
-	E_list *el1, *el2;
-
 	if ((e1 == NULL) && (e2 == NULL))
-	  return 1; /* they are equal */
+	  return true;
 	if ((e1 == NULL) || (e2 == NULL))
-	  return 0; /* they are not equal */
+	  return false;
 	if (e1->type != e2->type)
-		return 0;
-	if (fabs (e1->cost - e2->cost) > 0.001)
-		return 0;
+		return false;
+	if (!cost_eq(e1->cost, e2->cost))
+		return false;
+
 	if (e1->type == CONNECTOR_type)
 	{
+		if (e1->condesc != e2->condesc)
+			return false;
 		if (e1->dir != e2->dir)
-			return 0;
-		/* printf("%s %s\n",e1->u.condesc->string,e2->u.condesc->string); */
-		if (e1->u.condesc != e2->u.condesc)
-			return 0;
+			return false;
 	}
 	else
 	{
-		el1 = e1->u.l;
-		el2 = e2->u.l;
-		/* while at least 1 is non-null */
-		for (;(el1!=NULL)||(el2!=NULL);) {
-		   /*fail if 1 is null */
-			if ((el1==NULL)||(el2==NULL))
-				return 0;
-			/* fail if they are not compared */
-			if (!exp_compare(el1->e, el2->e))
-				return 0;
-			el1 = el1->next;
-			el2 = el2->next;
+		/* Iterate operands to avoid a deep recursion due to a lot of operands. */
+		for (e1 = e1->operand_first, e2 = e2->operand_first;
+		     (e1 != NULL) && (e2 != NULL);
+		     e1 = e1->operand_next, e2 = e2->operand_next)
+		{
+			if (!exp_compare(e1, e2))
+				return false;
 		}
+		return ((e1 == NULL) && (e2 == NULL));
 	}
-	return 1; /* if never returned 0, return 1 */
+	return true;
 }
 
 /**
@@ -150,12 +156,8 @@ static bool exp_compare(Exp * e1, Exp * e2)
  */
 static int exp_contains(Exp * super, Exp * sub)
 {
-	E_list * el;
-
 #if 0 /* DEBUG */
-	printf("SUP: ");
-	if (super) print_expression(super);
-	printf("\n");
+	if (super) printf("SUP: %s\n", lg_exp_stringify(super));
 #endif
 
 	if (sub==NULL || super==NULL)
@@ -167,8 +169,9 @@ static int exp_contains(Exp * super, Exp * sub)
 
 	/* proceed through supers children and return 1 if sub
 	   is contained in any of them */
-	for(el = super->u.l; el!=NULL; el=el->next) {
-		if (exp_contains(el->e, sub)==1)
+	for(Exp *opd = super->operand_first; opd != NULL; opd = opd->operand_next)
+	{
+		if (exp_contains(opd, sub)==1)
 			return 1;
 	}
 	return 0;
@@ -248,20 +251,19 @@ bool word_has_connector(Dict_node * dn, const char * cs, char direction)
 static bool exp_has_connector(const Exp * e, int depth, const char * cs,
                               char direction, bool smart_match)
 {
-	E_list * el;
 	if (e->type == CONNECTOR_type)
 	{
 		if (direction != e->dir) return false;
-		return smart_match ? easy_match(e->u.condesc->string, cs)
-		                   : string_set_cmp(e->u.condesc->string, cs);
+		return smart_match ? easy_match(e->condesc->string, cs)
+		                   : string_set_cmp(e->condesc->string, cs);
 	}
 
 	if (depth == 0) return false;
 	if (depth > 0) depth--;
 
-	for (el = e->u.l; el != NULL; el = el->next)
+	for (Exp *opd = e->operand_first; opd != NULL; opd = opd->operand_next)
 	{
-		if (exp_has_connector(el->e, depth, cs, direction, smart_match))
+		if (exp_has_connector(opd, depth, cs, direction, smart_match))
 			return true;
 	}
 	return false;
@@ -296,7 +298,7 @@ const char * word_only_connector(Dict_node * dn)
 {
 	Exp * e = dn->exp;
 	if (CONNECTOR_type == e->type)
-		return e->u.condesc->string;
+		return e->condesc->string;
 	return NULL;
 }
 
@@ -317,11 +319,8 @@ static bool dn_word_contains(Dictionary dict,
 	m_exp = m_dn->exp;
 
 #if 0 /* DEBUG */
-	printf("\nWORD: ");
-	print_expression(w_dn->exp);
-	printf("\nMACR: ");
-	print_expression(m_exp);
-	printf("\n");
+	printf("\nWORD: %s\n", lg_exp_stringify(w_dn->exp));
+	printf("\nMACR: %s\n", lg_exp_stringify(m_exp));
 #endif
 
 	for (;w_dn != NULL; w_dn = w_dn->right)
