@@ -67,6 +67,8 @@ static void dyn_ensure_empty_line(dyn_str *e)
 static void print_expression_tag_start(Dictionary dict, dyn_str *e, const Exp *n,
                                  int *indent)
 {
+	if (n->type == CONNECTOR_type) return;
+
 	switch (n->tag_type)
 	{
 		case Exptag_none:
@@ -92,6 +94,8 @@ static void print_expression_tag_start(Dictionary dict, dyn_str *e, const Exp *n
 static void print_expression_tag_end(Dictionary dict, dyn_str *e, const Exp *n,
                                  int *indent)
 {
+	if (n->type == CONNECTOR_type) return;
+
 	if (NULL == dict) return;
 
 	switch (n->tag_type)
@@ -262,7 +266,7 @@ typedef struct
 	int indent;
 	int pos;                 /* current connector position in expression */
 	int *find_pos;           /* disjunct expression connectors positions */
-	bool is_after_connector; /* an indicators for printing '&' */
+	bool is_after_connector; /* an indicator for printing '&' */
 } cmacro_context;
 
 /**
@@ -391,6 +395,8 @@ static const char *stringify_Exp_tag(Exp *e, Dictionary dict)
 {
 	static TLS char tag_info[64];
 
+	if (e->type == CONNECTOR_type) return "";
+
 		switch (e->tag_type)
 		{
 			case Exptag_none:
@@ -399,7 +405,7 @@ static const char *stringify_Exp_tag(Exp *e, Dictionary dict)
 				if (dict == NULL)
 				{
 					snprintf(tag_info, sizeof(tag_info), " dialect_tag=%u",
-					         e->tag_id);
+					         (unsigned int)e->tag_id);
 				}
 				else
 				{
@@ -482,7 +488,8 @@ void prt_exp_all(dyn_str *s, Exp *e, int i, Dictionary dict)
 	}
 }
 
-GNUC_UNUSED static void prt_exp_mem(Exp *e)
+#ifdef DEBUG
+GNUC_UNUSED void prt_exp_mem(Exp *e)
 {
 	dyn_str *s = dyn_str_new();
 
@@ -491,6 +498,7 @@ GNUC_UNUSED static void prt_exp_mem(Exp *e)
 	printf("%s", e_str);
 	free(e_str);
 }
+#endif
 
 /* ================ Print disjuncts and connectors ============== */
 static bool is_flag(uint32_t flags, char flag)
@@ -518,8 +526,8 @@ static uint32_t make_flags(const char *flags)
  * Optional [nearest_word, length_limit or farthest_word]: (flag 'l').
  * x: Shallow/deep indication as "s" / "d" (if shallow != -1)
  */
-static void dyn_print_one_connector(dyn_str *s, Connector *e, int dir,
-                                    int shallow, uint32_t flags)
+static void dyn_print_one_connector(dyn_str *s, const Connector *e, int dir,
+                                    uint32_t flags)
 {
 	if (e->multi)
 		dyn_strcat(s, "@");
@@ -530,48 +538,70 @@ static void dyn_print_one_connector(dyn_str *s, Connector *e, int dir,
 	if (is_flag(flags, 'r') && e->refcount)
 		append_string(s, "{%d}",e->refcount);
 	if (is_flag(flags, 'l'))
-		append_string(s, "(%d,%d)", e->nearest_word, e->length_limit);
+		append_string(s, "(%d,%d)", e->nearest_word, e->farthest_word);
 #if 0
 	append_string(s, "<<%d>>", e->exp_pos);
 #endif
-	if (-1 != shallow)
-		dyn_strcat(s, (0 == shallow) ? "d" : "s");
+	if (is_flag(flags, 's'))
+		dyn_strcat(s, e->shallow ? "s" : "d");
 }
 
-GNUC_UNUSED static void print_one_connector(Connector *e, int dir, int shallow,
-                                            uint32_t flags)
+void print_one_connector(const Connector *e, const char *flags)
 {
 	dyn_str *s = dyn_str_new();
+	int dir = -1;
 
-	dyn_print_one_connector(s, e, dir, shallow, flags);
+	if (flags == NULL) flags = "lt";
+	if (*flags == '-') { flags++; dir = 0; }
+	if (*flags == '+') { flags++; dir = 1; }
+
+	uint32_t int_flags = make_flags(flags);
+	dyn_print_one_connector(s, e, dir, int_flags);
 
 	char *t = dyn_str_take(s);
 	puts(t);
 	free(t);
 }
 
-static void dyn_print_connector_list(dyn_str *s, Connector *e, int dir,
+static void dyn_print_connector_list(dyn_str *s, const Connector *e, int dir,
                                      uint32_t flags)
 {
 
 	if (e == NULL) return;
 	dyn_print_connector_list(s, e->next, dir, flags);
 	if (e->next != NULL) dyn_strcat(s, " ");
-	dyn_print_one_connector(s, e, dir, /*shallow*/-1, flags);
+	dyn_print_one_connector(s, e, dir, flags);
 }
 
-void print_connector_list(Connector *e, const char *flags)
+/* Special flags here: Initial "-" or "+" for direction sign. */
+
+char *print_connector_list_str(const Connector *e, const char *flags)
 {
 	dyn_str *s = dyn_str_new();
+	int dir = -1;
 
 	if (flags == NULL) flags = "lt";
+	if (*flags == '-') { flags++; dir = 0; }
+	if (*flags == '+') { flags++; dir = 1; }
+
+	dyn_print_connector_list(s, e, dir, make_flags(flags));
+
+	return dyn_str_take(s);
+}
+
+char *print_one_connector_str(const Connector *e, const char *flags)
+{
+	dyn_str *s = dyn_str_new();
+	int dir = -1;
+
+	if (flags == NULL) flags = "lt";
+	if (*flags == '-') { flags++; dir = 0; }
+	if (*flags == '+') { flags++; dir = 1; }
+
 	uint32_t int_flags = make_flags(flags);
+	dyn_print_one_connector(s, e, dir, int_flags);
 
-	dyn_print_connector_list(s, e, /*dir*/-1, int_flags);
-
-	char *t = dyn_str_take(s);
-	puts(t);
-	free(t);
+	return dyn_str_take(s);
 }
 
 /* Ascending sort of connector positions. */
@@ -594,7 +624,8 @@ typedef struct
 	unsigned int num_tunnels;
 } select_data;
 
-static void dyn_print_disjunct_list(dyn_str *s, Disjunct *dj, uint32_t flags,
+static void dyn_print_disjunct_list(dyn_str *s, const Disjunct *dj,
+              uint32_t flags,
               bool (* select)(const char *dj_str, select_data *criterion),
               select_data *criterion)
 {
@@ -604,7 +635,16 @@ static void dyn_print_disjunct_list(dyn_str *s, Disjunct *dj, uint32_t flags,
 
 	for (;dj != NULL; dj=dj->next)
 	{
-		lg_strlcpy(word, dj->word_string, sizeof(word));
+		if (dj->is_category == 0)
+		{
+			lg_strlcpy(word, dj->word_string, sizeof(word));
+		}
+		else
+		{
+			int n = snprintf(word, sizeof(word), "%x", dj->category[0].num);
+			if (dj->num_categories > 1)
+				snprintf(word + n, sizeof(word) - n, " (%u)", dj->num_categories);
+		}
 		patch_subscript_mark(word);
 		dyn_str *l = dyn_str_new();
 
@@ -612,7 +652,8 @@ static void dyn_print_disjunct_list(dyn_str *s, Disjunct *dj, uint32_t flags,
 		if (print_disjunct_address) append_string(s, "(%p)", dj);
 		dyn_strcat(l, ": ");
 
-		append_string(l, "[%d]%s= ", djn++, cost_stringify(dj->cost));
+		const char *cost_str = cost_stringify(dj->cost);
+		append_string(l, "[%d]%s%s= ", djn++, &" "[(*cost_str == '-')], cost_str);
 		dyn_print_connector_list(l, dj->left, /*dir*/0, flags);
 		dyn_strcat(l, " <> ");
 		dyn_print_connector_list(l, dj->right, /*dir*/1, flags);
@@ -654,7 +695,7 @@ static void dyn_print_disjunct_list(dyn_str *s, Disjunct *dj, uint32_t flags,
 	}
 }
 
-void print_disjunct_list(Disjunct *d, const char *flags)
+void print_disjunct_list(const Disjunct *d, const char *flags)
 {
 	dyn_str *s = dyn_str_new();
 
@@ -739,11 +780,6 @@ static char *display_disjuncts(Dictionary dict, const Dict_node *dn,
 	                              /*num_elements*/65536, sizeof(Connector),
 	                              /*zero_out*/true, /*align*/false, false);
 
-	/* copy_Exp() needs an Exp memory pool. */
-	Pool_desc *Exp_pool = pool_new(__func__, "Exp", /*num_elements*/256,
-	                               sizeof(Exp), /*zero_out*/false,
-	                               /*align*/false, /*exact*/false);
-
 	select_data criterion = { .regex = rn };
 	void *select = (rn == NULL) ? NULL : select_disjunct;
 
@@ -752,12 +788,12 @@ static char *display_disjuncts(Dictionary dict, const Dict_node *dn,
 	for (; dn != NULL; dn = dn->right)
 	{
 		/* Use copy_Exp() to assign dialect cost. */
-		Exp *e = copy_Exp(dn->exp, Exp_pool, opts);
+		Exp *e = copy_Exp(dn->exp, dummy_sent->Exp_pool, opts);
 		Disjunct *d = build_disjuncts_for_exp(dummy_sent, e, dn->string, NULL,
 		                                      max_cost, NULL);
 
 		unsigned int dnum0 = count_disjuncts(d);
-		d = eliminate_duplicate_disjuncts(d);
+		d = eliminate_duplicate_disjuncts(d, NULL);
 		unsigned int dnum1 = count_disjuncts(d);
 
 		if ((flags != NULL) && (strchr(flags, 'm') != NULL))
@@ -770,7 +806,7 @@ static char *display_disjuncts(Dictionary dict, const Dict_node *dn,
 		dyn_print_disjunct_list(dyn_pdl, d, int_flags, select, &criterion);
 		char *dliststr = dyn_str_take(dyn_pdl);
 
-		pool_reuse(Exp_pool);
+		pool_reuse(dummy_sent->Exp_pool);
 		pool_reuse(dummy_sent->Disjunct_pool);
 		pool_reuse(dummy_sent->Connector_pool);
 
@@ -794,7 +830,6 @@ static char *display_disjuncts(Dictionary dict, const Dict_node *dn,
 				              criterion.num_selected == 1 ? "" : "s");
 		}
 	}
-	pool_delete(Exp_pool);
 	sentence_delete(dummy_sent);
 
 	return dyn_str_take(s);
@@ -806,8 +841,8 @@ static void notify_ignoring_flag(const char *flag)
 }
 
 /**
- * Copy \p characters from \p src to \p dst, while quoting with \c '\\'
- * the characters in \p q.
+ * Copy bytes from \p src to \p dst, while quoting with \c '\\'
+ * the bytes in \p q.
  * Note: This is not a general-purpose function, since:
  * - It assumes \p dst has enough space.
  * - \p src is not checked for NUL.
@@ -815,7 +850,7 @@ static void notify_ignoring_flag(const char *flag)
  *
  * @param src Source buffer.
  * @param dst[out] Destination buffer.
- * @return Number of characters add to \p dst.
+ * @return Number of bytes added to \p dst.
  */
 static size_t copy_quoted(const char *q, char *dst, const char *src, size_t len)
 {
@@ -859,8 +894,8 @@ static Regex_node *make_disjunct_pattern(const char *pattern, const char *flags)
 		notify_ignoring_flag(is_full);
 		notify_ignoring_flag(is_anyorder);
 
-		const char added_chars[] = "\\[]";
-		regpat = alloca(pat_len + sizeof(added_chars));
+		const size_t added_chars_len = sizeof("\\[]");
+		regpat = alloca(pat_len + added_chars_len);
 		strcpy(regpat, "\\[");
 		strcat(regpat, pattern);
 		strcat(regpat, "]");
@@ -894,8 +929,8 @@ static Regex_node *make_disjunct_pattern(const char *pattern, const char *flags)
 		}
 
 		/* Note: This section is sensitive to the disjunct print format. */
-		const char added_chars[] = "= <> $";
-		const size_t regpat_size = 2 * (pat_len + sizeof(added_chars));
+		const size_t added_chars_len = sizeof("= <> $");
+		const size_t regpat_size = 2 * (pat_len + added_chars_len);
 		regpat = alloca(regpat_size);
 
 		size_t dst_pos = lg_strlcpy(regpat, "= ", regpat_size);
@@ -939,7 +974,7 @@ static Regex_node *make_disjunct_pattern(const char *pattern, const char *flags)
 						alloca(strlen(constring) + sizeof(added_chars));
 					word_boundary_constring[0] = ' ';
 					strcpy(word_boundary_constring+1, constring);
-					strcat(word_boundary_constring+1, "( |$)");
+					strcat(word_boundary_constring+1, added_chars+1);
 					constring = word_boundary_constring;
 
 				}
@@ -1005,9 +1040,9 @@ static bool validate_flags(const char *display_type, const char *flags)
  * !!word*.sub and get a list of all words that match up to the wild-card.
  * In this case no split is done.
  *
- * @arg arg[0] specifies the display type. If it is equal to &do_display_expr
- * than this is request to display expressions. Else it is a request to
- * display disjuncts. arg[2] specifies the request flags.
+ * @param arg arg[0] specifies the display type. If it is equal to
+ * &do_display_expr than this is request to display expressions. Else it
+ * is a request to display disjuncts. arg[2] specifies the request flags.
  */
 static char *display_word_split(Dictionary dict,
                const char * word, Parse_Options opts,
@@ -1211,7 +1246,7 @@ static char *display_expr(Dictionary dict, const char *word, Dict_node *dn,
 /**
  * A callback function to display \p word number of disjuncts and file name.
  *
- * @arg Callback args (unused).
+ * @param arg Callback args (unused).
  * @return String to display. Must be freed by the caller.
  */
 static char *display_word_info(Dictionary dict, const char *word,
