@@ -12,15 +12,18 @@
 /*************************************************************************/
 
 #include "api-structures.h"
+#include "dict-common/dialect.h"      // dialect_alloc
 #include "dict-common/dict-affix.h"
 #include "dict-common/dict-api.h"
 #include "dict-common/dict-common.h"
+#include "dict-common/dict-defines.h"
 #include "dict-common/dict-impl.h"
 #include "dict-common/dict-utils.h"
 #include "dict-common/file-utils.h"
 #include "dict-common/idiom.h"
 #include "dict-common/regex-morph.h"
 #include "post-process/pp_knowledge.h"
+#include "read-dialect.h"
 #include "read-dict.h"
 #include "read-regex.h"
 #include "string-set.h"
@@ -106,6 +109,7 @@ dictionary_six_str(const char * lang,
 {
 	const char * t;
 	Dictionary dict;
+	size_t Exp_pool_size;
 
 	dict = (Dictionary) malloc(sizeof(struct Dictionary_s));
 	memset(dict, 0, sizeof(struct Dictionary_s));
@@ -120,8 +124,9 @@ dictionary_six_str(const char * lang,
 
 	if (NULL != affix_name)
 	{
-		/* To disable spell-checking, just set the checker to NULL */
-		dict->spell_checker = spellcheck_create(dict->lang);
+		if (!dictionary_generation_request(dict))
+			dict->spell_checker = spellcheck_create(dict->lang);
+
 #if defined HAVE_HUNSPELL || defined HAVE_ASPELL
 		/* FIXME: Move to spellcheck-*.c */
 		if (verbosity_level(D_USER_BASIC) && (NULL == dict->spell_checker))
@@ -135,7 +140,17 @@ dictionary_six_str(const char * lang,
 		dict->lookup_wild = file_lookup_wild;
 		dict->free_lookup = free_llist;
 		dict->lookup = file_boolean_lookup;
+		dict->dialect_tag.set = string_id_create();
 		condesc_init(dict, 1<<13);
+		Exp_pool_size = 1<<13;
+
+		if (!test_enabled("no-macro-tag"))
+		{
+			dict->macro_tag = malloc(sizeof(*dict->macro_tag));
+			memset(dict->macro_tag, 0, sizeof(*dict->macro_tag));
+		}
+
+		dict->define.set = string_id_create();
 	}
 	else
 	{
@@ -146,8 +161,13 @@ dictionary_six_str(const char * lang,
 		dict->insert_entry = load_affix;
 		dict->lookup = return_true;
 		condesc_init(dict, 1<<9);
+		Exp_pool_size = 1<<5;
 	}
 
+	dict->Exp_pool = pool_new(__func__, "Exp", /*num_elements*/Exp_pool_size,
+	                          sizeof(Exp), /*zero_out*/false,
+	                          /*align*/false, /*exact*/false);
+
 	/* Read dictionary from the input string. */
 
 	/* Make sure "." is used in string conversion to floating point.
@@ -171,7 +191,14 @@ dictionary_six_str(const char * lang,
 		return dict;
 	}
 
-	dictionary_setup_locale(dict);
+	if (dict->dialect_tag.num == 0)
+	{
+		string_id_delete(dict->dialect_tag.set);
+		dict->dialect_tag.set = NULL;
+	}
+
+	if (!dictionary_setup_defines(dict))
+		goto failure;
 
 	dict->affix_table = dictionary_six(lang, affix_name, NULL, NULL, NULL, NULL);
 	if (dict->affix_table == NULL)
@@ -206,14 +233,9 @@ dictionary_six_str(const char * lang,
 	locale = setlocale(LC_CTYPE, locale);            /* Restore the locale. */
 	assert(NULL != locale, "Cannot restore program locale");
 
-#ifdef USE_CORPUS
-	dict->corpus = lg_corpus_new();
-#endif
-
 	dict->base_knowledge  = pp_knowledge_open(pp_name);
 	dict->hpsg_knowledge  = pp_knowledge_open(cons_name);
 
-	dictionary_setup_defines(dict);
 	condesc_setup(dict);
 
 	// Special-case hack.
@@ -231,6 +253,7 @@ failure:
 /**
  * Use filenames of six different files to put together the dictionary.
  */
+NO_SAN_DICT
 Dictionary
 dictionary_six(const char * lang, const char * dict_name,
                const char * pp_name, const char * cons_name,
@@ -279,6 +302,24 @@ Dictionary dictionary_create_from_file(const char * lang)
 		free(cons_name);
 		free(pp_name);
 		free(dict_name);
+
+		if (dictionary != NULL)
+		{
+			char *dialect_name = join_path(lang, "4.0.dialect");
+			if (!dialect_file_read(dictionary, dialect_name))
+			{
+				dictionary_delete(dictionary);
+				dictionary = NULL;
+			}
+			else if ((dictionary->dialect == NULL) ||
+			         (dictionary->dialect->num_table_tags == 0))
+			{
+				/* No or empty dialect file. */
+				free_dialect(dictionary->dialect);
+				dictionary->dialect = NULL;
+			}
+			free(dialect_name);
+		}
 	}
 	else
 	{
