@@ -13,18 +13,19 @@
 
 #include <string.h>
 
-#include "api-structures.h"           // For Sentence_s (add_empty_word())
-#include "dict-common/dict-affix.h"   // For is_stem()
+#include "api-structures.h"             // Sentence_s (add_empty_word)
+#include "dict-common/dialect.h"
+#include "dict-common/dict-affix.h"     // is_stem
 #include "dict-common/dict-common.h"
-#include "dict-common/dict-defines.h" // For SUBSCRIPT_MARK
+#include "dict-common/dict-utils.h"     // patch_subscript
 #include "dict-common/file-utils.h"
 #include "dict-common/idiom.h"
 #include "error.h"
-#include "print/print.h"
 #include "externs.h"
+#include "print/print.h"
 #include "read-dict.h"
 #include "string-set.h"
-#include "tokenize/tok-structures.h" // needed for MT_WALL
+#include "tokenize/tok-structures.h"    // MT_WALL
 #include "utilities.h"
 #include "word-file.h"
 
@@ -42,19 +43,14 @@
   or "and" or "|" or "or", and there are three types of parentheses:
   "()", "{}", and "[]".  The terminal symbols of this grammar are the
   connectors, which are strings of letters or numbers or *s.
-  Expressions may be written in prefix or infix form. In prefix-form,
-  the expressions are lisp-like, with the operators &, | preceding
-  the operands. In infix-form, the operators are in the middle. The
-  current dictionaries are in infix form.  If the C preprocessor
-  constant INFIX_NOTATION is defined, then the dictionary is assumed
-  to be in infix form.
+  Expressions are written in infix form.
 
   The connector begins with an optional @, which is followed by an upper
   case sequence of letters. Each subsequent *, lower case letter or
   number is a subscript. At the end is a + or - sign.  The "@" allows
   this connector to attach to one or more other connectors.
 
-  Here is a sample dictionary entry (in infix form):
+  Here is a sample dictionary entry:
 
       gone:         T- & {@EV+};
 
@@ -137,7 +133,7 @@ static void dict_error2(Dictionary dict, const char * s, const char *s2)
 	int save_line_number    = dict->line_number;
 
 	tokens[0] = '\0';
-	for (i=0; i<5 && dict->token[0] != '\0' ; i++)
+	for (i=0; i<5 && dict->token[0] != '\0'; i++)
 	{
 		pos += snprintf(t, ERRBUFLEN, "\"%s\" ", dict->token);
 		strncat(tokens, t, ERRBUFLEN-1-pos);
@@ -631,7 +627,7 @@ static bool subscr_match(const char *s, const Dict_node * dn)
 	if (NULL == s_sub) return true;
 	t_sub = strrchr(dn->string, SUBSCRIPT_MARK);
 	if (NULL == t_sub) return false;
-	if ( 0 == strcmp(s_sub, t_sub)) return true;
+	if (0 == strcmp(s_sub, t_sub)) return true;
 
 	return false;
 }
@@ -644,7 +640,7 @@ static bool subscr_match(const char *s, const Dict_node * dn)
  */
 static Dict_node *
 rdictionary_lookup(Dict_node * restrict llist,
-                   const Dict_node * restrict dn,
+                   Dict_node * restrict dn,
                    const char * restrict s,
                    bool match_idiom,
                    int (*dict_order)(const char *, const Dict_node *))
@@ -666,6 +662,7 @@ rdictionary_lookup(Dict_node * restrict llist,
 		dn_new = dict_node_new();
 		*dn_new = *dn;
 		dn_new->right = llist;
+		dn_new->left = dn; /* Currently only used for inserting idioms */
 		llist = dn_new;
 	}
 	if (m <= 0)
@@ -794,26 +791,40 @@ static Dict_node * strict_lookup_list(const Dictionary dict, const char *s)
 
 /* ======================================================================== */
 /**
- * Allocate a new Exp node and link it into the exp_list for freeing later.
+ * Allocate a new Exp node.
  */
-Exp * Exp_create(Exp_list *eli)
+Exp *Exp_create(Pool_desc *mp)
 {
-	Exp * e = malloc(sizeof(Exp));
-	e->next = eli->exp_list;
-	eli->exp_list = e;
+	Exp *e = pool_alloc(mp);
+	e->tag_type = Exptag_none;
 	return e;
 }
 
+/**
+ * Duplicate the given Exp node.
+ * This is needed in case it participates more than once in a
+ * single expression.
+ */
+Exp *Exp_create_dup(Pool_desc *mp, Exp *old_e)
+{
+	Exp *new_e = Exp_create(mp);
+
+	*new_e = *old_e;
+
+	return new_e;
+}
+
 /**
  * This creates a node with zero children.  Initializes
  * the cost to zero.
  */
-static Exp * make_zeroary_node(Exp_list * eli)
+static Exp * make_zeroary_node(Pool_desc *mp)
 {
-	Exp * n = Exp_create(eli);
+	Exp * n = Exp_create(mp);
 	n->type = AND_type;  /* these must be AND types */
 	n->cost = 0.0;
-	n->u.l = NULL;
+	n->operand_first = NULL;
+	n->operand_next = NULL;
 	return n;
 }
 
@@ -821,15 +832,14 @@ static Exp * make_zeroary_node(Exp_list * eli)
  * This creates a node with one child (namely e).  Initializes
  * the cost to zero.
  */
-static Exp * make_unary_node(Exp_list * eli, Exp * e)
+Exp *make_unary_node(Pool_desc *mp, Exp * e)
 {
 	Exp * n;
-	n = Exp_create(eli);
+	n = Exp_create(mp);
 	n->type = AND_type;  /* these must be AND types */
+	n->operand_next = NULL;
 	n->cost = 0.0;
-	n->u.l = (E_list *) malloc(sizeof(E_list));
-	n->u.l->next = NULL;
-	n->u.l->e = e;
+	n->operand_first = e;
 	return n;
 }
 
@@ -837,63 +847,50 @@ static Exp * make_unary_node(Exp_list * eli, Exp * e)
  * Create an AND_type expression. The expressions nl, nr will be
  * AND-ed together.
  */
-static Exp * make_and_node(Exp_list * eli, Exp* nl, Exp* nr)
+static Exp * make_and_node(Pool_desc *mp, Exp* nl, Exp* nr)
 {
-	E_list *ell, *elr;
 	Exp* n;
 
-	n = Exp_create(eli);
+	n = Exp_create(mp);
 	n->type = AND_type;
+	n->operand_next = NULL;
 	n->cost = 0.0;
 
-	n->u.l = ell = (E_list *) malloc(sizeof(E_list));
-	ell->next = elr = (E_list *) malloc(sizeof(E_list));
-	elr->next = NULL;
+	n->operand_first = nl;
+	nl->operand_next = nr;
+	nr->operand_next = NULL;
 
-	ell->e = nl;
-	elr->e = nr;
 	return n;
 }
 
-static Exp *make_op_Exp(Exp_list *eli, Exp_type t)
+static Exp *make_op_Exp(Pool_desc *mp, Exp_type t)
 {
-	Exp * n = Exp_create(eli);
+	Exp * n = Exp_create(mp);
 	n->type = t;
+	n->operand_next = NULL;
 	n->cost = 0.0;
-	n->u.l = NULL;
 
+	/* The caller is supposed to assign n->operand. */
 	return n;
 }
 
-static E_list *make_E_list_val(Exp* nr)
-{
-	E_list *elr = (E_list *) malloc(sizeof(E_list));
-
-	elr->e = nr;
-	elr->next = NULL;
-
-	return elr;
-}
-
 /**
  * Create an OR_type expression. The expressions nl, nr will be
  * OR-ed together.
  */
-static Exp * make_or_node(Exp_list *eli, Exp* nl, Exp* nr)
+static Exp * make_or_node(Pool_desc *mp, Exp* nl, Exp* nr)
 {
-	E_list *ell, *elr;
 	Exp* n;
 
-	n = Exp_create(eli);
+	n = Exp_create(mp);
 	n->type = OR_type;
+	n->operand_next = NULL;
 	n->cost = 0.0;
 
-	n->u.l = ell = (E_list *) malloc(sizeof(E_list));
-	ell->next = elr = (E_list *) malloc(sizeof(E_list));
-	elr->next = NULL;
+	n->operand_first = nl;
+	nl->operand_next = nr;
+	nr->operand_next = NULL;
 
-	ell->e = nl;
-	elr->e = nr;
 	return n;
 }
 
@@ -902,9 +899,9 @@ static Exp * make_or_node(Exp_list *eli, Exp* nl, Exp* nr)
  * and the other as zeroary node.  This has the effect of creating
  * what used to be called an optional node.
  */
-static Exp * make_optional_node(Exp_list *eli, Exp * e)
+static Exp *make_optional_node(Pool_desc *mp, Exp *e)
 {
-	return make_or_node(eli, make_zeroary_node(eli), e);
+	return make_or_node(mp, make_zeroary_node(mp), e);
 }
 
 /**
@@ -915,7 +912,7 @@ static Exp * make_optional_node(Exp_list *eli, Exp * e)
  */
 static Exp * make_dir_connector(Dictionary dict, int i)
 {
-	Exp* n = Exp_create(&dict->exp_list);
+	Exp* n = Exp_create(dict->Exp_pool);
 	char *constring;
 
 	n->dir = dict->token[i];
@@ -931,15 +928,39 @@ static Exp * make_dir_connector(Dictionary dict, int i)
 		n->multi = false;
 	}
 
-	n->u.condesc = condesc_add(&dict->contable,
-	                            string_set_add(constring, dict->string_set));
-	if (NULL == n->u.condesc) return NULL; /* Table ovf */
+	n->condesc = condesc_add(&dict->contable,
+	                         string_set_add(constring, dict->string_set));
+	if (NULL == n->condesc) return NULL; /* Table ovf */
 	n->type = CONNECTOR_type;
+	n->operand_next = NULL; /* unused, but accessed by copy_Exp() and some more */
 	n->cost = 0.0;
 	return n;
 }
 
 /* ======================================================================== */
+/**
+ * Add an optional macro/word tag, for expression debugging.
+ * Enabled by !test="macro-tag". This tag is used only in expression printing.
+ */
+static unsigned int exptag_macro_add(Dictionary dict, const char *tag)
+{
+	expression_tag *mt = dict->macro_tag;
+	if (mt == NULL) return 0;
+
+	if (mt->num == mt->size)
+	{
+		if (mt->num == 0)
+			mt->size = 128;
+		else
+			mt->size *= 2;
+
+		mt->name = realloc(mt->name, mt->size * sizeof(*mt->name));
+	}
+	mt->name[mt->num] = tag;
+
+	return mt->num++;
+}
+
 /**
  * make_connector() -- make a node for a connector or dictionary word.
  *
@@ -973,8 +994,10 @@ static Exp * make_connector(Dictionary dict)
 			return NULL;
 		}
 
-		/* Wrap it in a unary node as a placeholder for a cost if needed. */
-		n = make_unary_node(&dict->exp_list, dn->exp);
+		/* Wrap it in a unary node as a placeholder for a macro tag and cost. */
+		n = make_unary_node(dict->Exp_pool, dn->exp);
+		n->tag_id = exptag_macro_add(dict, dn->string);
+		if (n->tag_id != 0) n->tag_type = Exptag_macro;
 
 		file_free_lookup(dn_head);
 	}
@@ -1003,7 +1026,7 @@ static Exp * make_connector(Dictionary dict)
 			min = make_dir_connector(dict, i);
 			if (NULL == min) return NULL;
 
-			n = make_or_node(&dict->exp_list, plu, min);
+			n = make_or_node(dict->Exp_pool, plu, min);
 		}
 		else
 		{
@@ -1023,42 +1046,6 @@ static Exp * make_connector(Dictionary dict)
 /* ======================================================================== */
 /* Empty-word handling. */
 
-/** Expression-creating function versions that use the Sentence
- *  expression memory pools.
- *  (See the description of the similar functions above.)
- */
-static Exp *make_or_node_from_pool(Sentence sent, Exp *nl, Exp *nr)
-{
-	E_list *ell, *elr;
-	Exp* n;
-
-	n = pool_alloc(sent->Exp_pool);
-	n->type = OR_type;
-	n->cost = 0.0;
-
-	n->u.l = ell = (E_list *) pool_alloc(sent->E_list_pool);
-	ell->next = elr = (E_list *) pool_alloc(sent->E_list_pool);
-	elr->next = NULL;
-
-	ell->e = nl;
-	elr->e = nr;
-	return n;
-}
-
-static Exp *make_zeroary_node_from_pool(Sentence sent)
-{
-	Exp * n = pool_alloc(sent->Exp_pool);
-	n->type = AND_type;  /* these must be AND types */
-	n->cost = 0.0;
-	n->u.l = NULL;
-	return n;
-}
-
-static Exp *make_optional_node_from_pool(Sentence sent, Exp *e)
-{
-	return make_or_node_from_pool(sent, make_zeroary_node_from_pool(sent), e);
-}
-
 /** Insert ZZZ+ connectors.
  *  This function was mainly used to support using empty-words, a concept
  *  that has been eliminated. However, it is still used to support linking of
@@ -1067,7 +1054,6 @@ static Exp *make_optional_node_from_pool(Sentence sent, Exp *e)
 void add_empty_word(Sentence sent, X_node *x)
 {
 	Exp *zn, *an;
-	E_list *elist, *flist;
 	const char *ZZZ = string_set_lookup(EMPTY_CONNECTOR, sent->dict->string_set);
 	/* This function is called only if ZZZ is in the dictionary. */
 
@@ -1083,29 +1069,22 @@ void add_empty_word(Sentence sent, X_node *x)
 		//lgdebug(+0, "Processing '%s'\n", x->string);
 
 		/* zn points at {ZZZ+} */
-		zn = pool_alloc(sent->Exp_pool);
+		zn = Exp_create(sent->Exp_pool);
 		zn->dir = '+';
-		zn->u.condesc = condesc_add(&sent->dict->contable, ZZZ);
+		zn->condesc = condesc_add(&sent->dict->contable, ZZZ);
 		zn->multi = false;
 		zn->type = CONNECTOR_type;
+		zn->operand_next = NULL; /* unused, but to be on the safe side */
 		zn->cost = 0.0;
-		zn = make_optional_node_from_pool(sent, zn);
-
-		/* flist is plain-word-exp */
-		flist = (E_list *) pool_alloc(sent->E_list_pool);
-		flist->next = NULL;
-		flist->e = x->exp;
-
-		/* elist is {ZZZ+} , (plain-word-exp) */
-		elist = (E_list *) pool_alloc(sent->E_list_pool);
-		elist->next = flist;
-		elist->e = zn;
+		zn = make_optional_node(sent->Exp_pool, zn);
 
 		/* an will be {ZZZ+} & (plain-word-exp) */
-		an = pool_alloc(sent->Exp_pool);
+		an = Exp_create(sent->Exp_pool);
 		an->type = AND_type;
+		an->operand_next = NULL;
 		an->cost = 0.0;
-		an->u.l = elist;
+		an->operand_first = zn;
+		zn->operand_next = x->exp;
 
 		x->exp = an;
 	}
@@ -1119,6 +1098,7 @@ void add_empty_word(Sentence sent, X_node *x)
  */
 static bool is_number(const char * str)
 {
+	if (str[0] == '\0') return false; /* End of file. */
 	if ('+' == str[0] || '-' == str[0]) str++;
 	size_t numlen = strspn(str, "0123456789.");
 
@@ -1127,139 +1107,16 @@ static bool is_number(const char * str)
 
 /* ======================================================================== */
 
-/* INFIX_NOTATION is always defined; we simply never use the format below. */
-/* #if ! defined INFIX_NOTATION */
-#if 0
-
-static Exp * expression(Dictionary dict);
-/**
- * We're looking at the first of the stuff after an "and" or "or".
- * Build a Exp node for this expression.  Set the cost and optional
- * fields to the default values.  Set the type field according to type
- */
-static Exp * operator_exp(Dictionary dict, int type)
-{
-	Exp * n;
-	E_list first;
-	E_list * elist;
-	n = Exp_create(dict);
-	n->type = type;
-	n->cost = 0.0;
-	elist = &first;
-	while((!is_equal(dict, ')')) && (!is_equal(dict, ']')) && (!is_equal(dict, '}'))) {
-		elist->next = (E_list *) malloc(sizeof(E_list));
-		elist = elist->next;
-		elist->next = NULL;
-		elist->e = expression(dict);
-		if (elist->e == NULL) {
-			return NULL;
-		}
-	}
-	if (elist == &first) {
-		dict_error(dict, "An \"or\" or \"and\" of nothing");
-		return NULL;
-	}
-	n->u.l = first.next;
-	return n;
-}
-
-/**
- * Looks for the stuff that is allowed to be inside of parentheses
- * either & or | followed by a list, or a terminal symbol.
- */
-static Exp * in_parens(Dictionary dict)
-{
-	Exp * e;
-
-	if (is_equal(dict, '&') || (strcmp(token, "and")==0)) {
-		if (!link_advance(dict)) {
-			return NULL;
-		}
-		return operator_exp(dict, AND_type);
-	} else if (is_equal(dict, '|') || (strcmp(dict->token, "or")==0)) {
-		if (!link_advance(dict)) {
-			return NULL;
-		}
-		return operator_exp(dict, OR_type);
-	} else {
-		return expression(dict);
-	}
-}
-
 /**
  * Build (and return the root of) the tree for the expression beginning
  * with the current token.  At the end, the token is the first one not
  * part of this expression.
  */
-static Exp * expression(Dictionary dict)
-{
-	Exp * n;
-	if (is_equal(dict, '(')) {
-		if (!link_advance(dict)) {
-			return NULL;
-		}
-		n = in_parens(dict);
-		if (!is_equal(dict, ')')) {
-			dict_error(dict, "Expecting a \")\".");
-			return NULL;
-		}
-		if (!link_advance(dict)) {
-			return NULL;
-		}
-	} else if (is_equal(dict, '{')) {
-		if (!link_advance(dict)) {
-			return NULL;
-		}
-		n = in_parens(dict);
-		if (!is_equal(dict, '}')) {
-			dict_error(dict, "Expecting a \"}\".");
-			return NULL;
-		}
-		if (!link_advance(dict)) {
-			return NULL;
-		}
-		n = make_optional_node(dict, n);
-	} else if (is_equal(dict, '[')) {
-		if (!link_advance(dict)) {
-			return NULL;
-		}
-		n = in_parens(dict);
-		if (!is_equal(dict, ']')) {
-			dict_error(dict, "Expecting a \"]\".");
-			return NULL;
-		}
-		if (!link_advance(dict)) {
-			return NULL;
-		}
-		n->cost += 1.0;
-	} else if (!dict->is_special) {
-		n = make_connector(dict);
-		if (n == NULL) {
-			return NULL;
-		}
-	} else if (is_equal(dict, ')') || is_equal(dict, ']')) {
-		/* allows "()" or "[]" */
-		n = make_zeroary_node(dict);
-	} else {
-			dict_error(dict, "Connector, \"(\", \"[\", or \"{\" expected.");
-		return NULL;
-	}
-	return n;
-}
-
-/* ======================================================================== */
-#else /* This is for infix notation */
-
-/**
- * Build (and return the root of) the tree for the expression beginning
- * with the current token.  At the end, the token is the first one not
- * part of this expression.
- */
-static Exp * make_expression(Dictionary dict)
+static Exp *make_expression(Dictionary dict)
 {
 	Exp *nl = NULL;
 	Exp *e_head = NULL;
-	E_list *el_tail = NULL; /* last part of the expression */
+	Exp *e_tail = NULL; /* last part of the expression */
 	bool is_sym_and = false;
 
 	while (true)
@@ -1297,7 +1154,7 @@ static Exp * make_expression(Dictionary dict)
 			if (!link_advance(dict)) {
 				return NULL;
 			}
-			nl = make_optional_node(&dict->exp_list, nl);
+			nl = make_optional_node(dict->Exp_pool, nl);
 		}
 		else if (is_equal(dict, '['))
 		{
@@ -1316,14 +1173,15 @@ static Exp * make_expression(Dictionary dict)
 				return NULL;
 			}
 
-			/* A square bracket can have a number after it.
-			 * If it is present, then that number is interpreted
-			 * as the cost of the bracket. Else, the cost of a
+			/* A square bracket can have a number or a name after it.
+			 * If a number is present, then that it is interpreted
+			 * as the cost of the bracket. If a name is present, it
+			 * is used as an expression tag. Else, the cost of a
 			 * square bracket is 1.0.
 			 */
 			if (is_number(dict->token))
 			{
-				double cost;
+				float cost;
 
 				if (strtodC(dict->token, &cost))
 				{
@@ -1338,6 +1196,28 @@ static Exp * make_expression(Dictionary dict)
 					return NULL;
 				}
 			}
+			else if ((strcmp(dict->token, "or") != 0) &&
+			         (strcmp(dict->token, "and") != 0) &&
+			         isalpha(dict->token[0]))
+			{
+				const char *bad = valid_dialect_name(dict->token);
+				if (bad != NULL)
+				{
+					char badchar[] = { *bad, '\0' };
+					dict_error2(dict, "Invalid character in dialect tag name:",
+					           badchar);
+					return NULL;
+				}
+				if (nl->tag_type != Exptag_none)
+				{
+					nl = make_unary_node(dict->Exp_pool, nl);
+				}
+				nl->tag_id = exptag_dialect_add(dict, dict->token);
+				nl->tag_type = Exptag_dialect;
+				if (!link_advance(dict)) {
+					return NULL;
+				}
+			}
 			else
 			{
 				nl->cost += 1.0;
@@ -1353,7 +1233,7 @@ static Exp * make_expression(Dictionary dict)
 		else if (is_equal(dict, ')') || is_equal(dict, ']'))
 		{
 			/* allows "()" or "[]" */
-			nl = make_zeroary_node(&dict->exp_list);
+			nl = make_zeroary_node(dict->Exp_pool);
 		}
 		else
 		{
@@ -1363,34 +1243,29 @@ static Exp * make_expression(Dictionary dict)
 
 		if (is_sym_and)
 		{
-			/* Part 2/2 of SYM_AND processing */
-			E_list **elp; /* LHS argument and the SYM_AND result placeholder */
-			Exp *nr, *na, *nb, *or;
+			/* Part 2/2 of SYM_AND processing. */
 
-			elp = (el_tail == NULL) ? &e_head->u.l : &el_tail;
-			nr = nl; /* last expression is actually the RHS */
-			nl = (*elp)->e;
+			/* Expand A ^ B into the expr ((A & B) or (B & A)). */
+			Exp *na = make_and_node(dict->Exp_pool,
+			                   Exp_create_dup(dict->Exp_pool, e_tail),
+			                   Exp_create_dup(dict->Exp_pool, nl));
+			Exp *nb = make_and_node(dict->Exp_pool,
+			                   Exp_create_dup(dict->Exp_pool, nl),
+			                   Exp_create_dup(dict->Exp_pool, e_tail));
+			Exp *or = make_or_node(dict->Exp_pool, na, nb);
 
-			/* Expand A ^ B into the expr ((A & B) or (B & A)).
-			 * It can be mixed with ordinary ands at the same level. */
-			na = make_and_node(&dict->exp_list, nl, nr);
-			nb = make_and_node(&dict->exp_list, nr, nl);
-			or = make_or_node(&dict->exp_list, na, nb);
-
-			(*elp)->e = or;
+			*e_tail = *or; /* SYM_AND result */
 			is_sym_and = false;
 		}
-		else if (el_tail != NULL)
+		else if (e_tail != NULL)
 		{
 			/* This is Non-commuting AND or Commuting OR.
-			 * Append the just read expression (nl) to its E_list chain. */
-			el_tail->next = make_E_list_val(nl);
-			el_tail = el_tail->next;
+			 * Append the just read expression (nl) to its operand chain. */
+			e_tail->operand_next = nl;
+			e_tail = nl;
 		}
 
-		/* If (el_tail == NULL), handle_basic_op() below will create an
-		 * E_list element with nl. Else it will just validate that the
-		 * current operation is consistent with the current expression level. */
+		/* Extract the operator. */
 
 		Exp_type op;
 
@@ -1408,7 +1283,7 @@ static Exp * make_expression(Dictionary dict)
 		else if (is_equal(dict, SYM_AND) || (strcmp(dict->token, "sym") == 0))
 		{
 			/* Part 1/2 of SYM_AND processing */
-			op = AND_type;
+			op = AND_type; /* allow mixing with ordinary ands at the same level */
 			is_sym_and = true; /* processing to be completed after next argument */
 		}
 		else
@@ -1417,7 +1292,15 @@ static Exp * make_expression(Dictionary dict)
 			return nl;
 		}
 
-		if (e_head != NULL)
+		/* If this is the first operand, use nl for it.  Else just validate
+		 * that the current operator is consistent with that of the current
+		 * expression level. */
+		if (e_head == NULL)
+		{
+			e_head = make_op_Exp(dict->Exp_pool, op);
+			e_head->operand_first = nl;
+		}
+		else
 		{
 			if (e_head->type != op)
 			{
@@ -1425,24 +1308,17 @@ static Exp * make_expression(Dictionary dict)
 				return NULL;
 			}
 		}
-		else
-		{
-			e_head = make_op_Exp(&dict->exp_list, op);
-			e_head->u.l = make_E_list_val(nl);
-		}
 
 		if (!link_advance(dict)) {
 			return NULL;
 		}
 
-		if (el_tail == NULL)
-			el_tail = e_head->u.l;
+		if (e_tail == NULL)
+			e_tail = e_head->operand_first;
 	}
 		/* unreachable */
 }
 
-#endif
-
 /* ======================================================================== */
 /* Implementation of the DSW algo for rebalancing a binary tree.
  * The point is -- after building the dictionary tree, we rebalance it
@@ -1531,7 +1407,7 @@ static Dict_node * dsw_vine_to_tree (Dict_node *root, int size)
 	vine_head.right = root;
 
 	dsw_compression(&vine_head, size - full_count);
-	for (size = full_count ; size > 1 ; size /= 2)
+	for (size = full_count; size > 1; size /= 2)
 	{
 		dsw_compression(&vine_head, size / 2);
 	}
@@ -1549,12 +1425,18 @@ static Dict_node * dsw_vine_to_tree (Dict_node *root, int size)
  * before being used.  The DSW algo below is ideal for that.
  */
 NO_SAN_DICT
-Dict_node * insert_dict(Dictionary dict, Dict_node * n, Dict_node * newnode)
+Dict_node *insert_dict(Dictionary dict, Dict_node *n, Dict_node *newnode)
 {
 	if (NULL == n) return newnode;
 
-	static Exp null_exp = { .type = AND_type, .u.l = NULL };
+	static Exp null_exp =
+	{
+		.type = AND_type,
+		.operand_first = NULL,
+		.operand_next = NULL,
+	};
 	int comp = dict_order_strict(newnode->string, n);
+
 	if (0 == comp &&
 	    /* Suppress reporting duplicate idioms until they are fixed. */
 	    (!contains_underbar(newnode->string) || test_enabled("dup-idioms")))
@@ -1691,7 +1573,7 @@ void insert_list(Dictionary dict, Dict_node * p, int l)
 	if (is_idiom_word(dn->string))
 	{
 		prt_error("Warning: Word \"%s\" found near line %d of \"%s\".\n"
-		        "\tWords ending \".Ix\" (x a number) are reserved for idioms.\n"
+		        "\tWords ending \".I\" are reserved for idioms.\n"
 		        "\tThis word will be ignored.\n",
 		        dn->string, dict->line_number, dict->name);
 		free(dn);
@@ -1750,6 +1632,94 @@ void insert_list(Dictionary dict, Dict_node * p, int l)
 	insert_list(dict, dn_second_half, l-k-1);
 }
 
+static void add_define(Dictionary dict, const char *name, const char *value)
+{
+	int id = string_id_add(name, dict->define.set);
+
+	if (!IS_DB_DICT(dict) && (dict->define.size >= (unsigned int)id))
+	{
+		prt_error("Warning: Redefinition of \"%s\", "
+		          "found near line %d of \"%s\"\n",
+		          name, dict->line_number, dict->name);
+	}
+	else
+	{
+		dict->define.size++;
+		dict->define.value =
+			realloc(dict->define.value, dict->define.size * sizeof(char *));
+		dict->define.name =
+			realloc(dict->define.name, dict->define.size * sizeof(char *));
+		assert(dict->define.size == (unsigned int)id,
+		       "\"define\" array size inconsistency");
+		dict->define.name[id - 1] = string_set_add(name, dict->string_set);
+	}
+	dict->define.value[id - 1] = string_set_add(value, dict->string_set);
+}
+
+static bool is_directive(const char *s)
+{
+	return
+		(strcmp(s, UNLIMITED_CONNECTORS_WORD) == 0) ||
+		(strncmp(s, LIMITED_CONNECTORS_WORD, sizeof(LIMITED_CONNECTORS_WORD)-1) == 0);
+}
+
+static bool is_correction(const char *s)
+{
+	static const char correction_mark[] = { SUBSCRIPT_MARK, '#' , '\0'};
+	return strstr(s, correction_mark) != 0;
+}
+
+static void add_category(Dictionary dict, Exp *e, Dict_node *dn, int n)
+{
+	if (n == 1)
+	{
+		if (is_macro(dn->string)) return;
+		if (!dict->generate_walls && is_wall(dn->string)) return;
+		if (is_correction(dn->string)) return;
+		if (is_directive(dn->string)) return;
+	}
+
+	/* Add a category with a place for n words. */
+	dict->num_categories++;
+	if (dict->num_categories >= dict->num_categories_alloced)
+	{
+		dict->num_categories_alloced *= 2;
+		dict->category =
+			realloc(dict->category,
+			        sizeof(*dict->category) * dict->num_categories_alloced);
+	}
+	dict->category[dict->num_categories].word =
+		malloc(sizeof(dict->category[0].word) * n);
+
+	n = 0;
+	for (Dict_node *dnx = dn; dnx != NULL; dnx = dnx->left)
+	{
+		if (is_macro(dnx->string)) continue;
+		if (!dict->generate_walls && is_wall(dnx->string)) continue;
+		if (is_correction(dnx->string)) continue;
+		if (is_directive(dnx->string)) return;
+		dict->category[dict->num_categories].word[n] = dnx->string;
+		n++;
+	}
+
+	if (n == 0)
+	{
+		free(dict->category[dict->num_categories].word);
+		--dict->num_categories;
+	}
+	else
+	{
+		assert(dict->num_categories < 1024 * 1024, "Insane number of categories");
+		char category_string[16]; /* For the tokenizer - not used here */
+		snprintf(category_string, sizeof(category_string), " %x",
+		         dict->num_categories);
+		string_set_add(category_string, dict->string_set);
+		dict->category[dict->num_categories].exp = e;
+		dict->category[dict->num_categories].num_words = n;
+		dict->category[dict->num_categories].name = "";
+	}
+}
+
 /**
  * read_entry() -- read one dictionary entry
  * Starting with the current token, parse one dictionary entry.
@@ -1848,6 +1818,22 @@ static bool read_entry(Dictionary dict)
 
 			return true;
 		}
+		else if (0 == strcmp(dict->token, "#define"))
+		{
+			if (!link_advance(dict)) goto syntax_error;
+			const char *name = strdupa(dict->token);
+
+			/* Get the value. */
+			if (!link_advance(dict)) goto syntax_error;
+			add_define(dict, name, dict->token);
+
+			if (!link_advance(dict)) goto syntax_error;
+			if (!is_equal(dict, ';'))
+			{
+				dict_error(dict, "Expecting \";\" at the end of #define.");
+				goto syntax_error;
+			}
+		}
 		else
 		{
 			Dict_node * dn_new = dict_node_new();
@@ -1885,14 +1871,23 @@ static bool read_entry(Dictionary dict)
 		goto syntax_error;
 	}
 
+	if (dn == NULL)
+	{
+		dict_error(dict, "Expecting a token before \":\".");
+		goto syntax_error;
+	}
+
 	/* At this point, dn points to a list of Dict_nodes connected by
-	 * their left pointers. These are to be inserted into the dictionary */
+	 * their left pointers. These are to be inserted into the dictionary. */
 	i = 0;
 	for (dnx = dn; dnx != NULL; dnx = dnx->left)
 	{
 		dnx->exp = n;
 		i++;
 	}
+	if (IS_GENERATION(dict))
+		add_category(dict, n, dn, i);
+
 	dict->insert_entry(dict, dn, i);
 
 	if (dict->suppress_warning)
@@ -1915,14 +1910,22 @@ syntax_error:
 	return false;
 }
 
+void print_dictionary_defines(Dictionary dict)
+{
+	for (size_t i = 0; i < dict->define.size; i++)
+	{
+		const char *value = dict->define.value[i];
+		int q = (int)(strcspn(value, SPECIAL) == strlen(value));
+		printf("#define %s %s%s%s\n",
+		       dict->define.name[i], &"\""[q], value, &"\""[q]);
+	}
+}
 
 static void rprint_dictionary_data(Dict_node * n)
 {
 	if (n == NULL) return;
 	rprint_dictionary_data(n->left);
-	printf("%s: ", n->string);
-	print_expression(n->exp);
-	printf("-6-\n");
+	printf("%s: %s\n", n->string, lg_exp_stringify(n->exp));
 	rprint_dictionary_data(n->right);
 }
 
@@ -1953,6 +1956,18 @@ bool read_dictionary(Dictionary dict)
 			return false;
 		}
 	}
+
+	if (dict->category != NULL)
+	{
+		/* Create a category element which contains 0 words, to signify the
+		 * end of the category array for the user API. The number of
+		 * categories will not get changed because macros are considered an
+		 * invalid word. */
+		Exp dummy_exp;
+		add_category(dict, &dummy_exp, NULL, 0);
+		dict->category[dict->num_categories + 1].num_words = 0;
+	}
+
 	dict->root = dsw_tree_to_vine(dict->root);
 	dict->root = dsw_vine_to_tree(dict->root, dict->num_entries);
 	return true;
