@@ -14,6 +14,7 @@
 
 #include "api-structures.h"             // Sentence
 #include "connectors.h"
+#include "dict-common/dict-api.h"
 #include "dict-common/dict-structures.h"
 #include "dict-common/dict-utils.h"     // copy_Exp
 #include "dict-common/regex-morph.h"    // match_regex
@@ -21,11 +22,93 @@
 #include "memory-pool.h"
 #include "prepare/build-disjuncts.h"
 #include "print/print-util.h"
+#include "string-set.h"
 #include "tokenize/tok-structures.h"    // XXX TODO provide gword access methods!
 #include "tokenize/word-structures.h"
 #include "tracon-set.h"
 #include "utilities.h"
 
+/* Disjunct API ... */
+
+static char *connector_list_to_expression(const char *connector_list)
+{
+	dyn_str *e = dyn_str_new();
+	for (const char *p = connector_list; *p != '\0'; p++)
+	{
+		if (*p != ' ')
+		{
+			dyn_strcat(e, (char []){ *p, '\0' });
+			continue;
+		}
+		if (p[1] != '\0') dyn_strcat(e, " & ");
+	}
+
+	return dyn_str_take(e);
+}
+
+/**
+ * Return the expression of the given disjunct;
+ * The caller has to free the returned value.
+ */
+char * disjunct_expression(Disjunct *d)
+{
+	char *ls = print_connector_list_str(d->left, "-");
+	char *rs = print_connector_list_str(d->right, "+");
+
+	size_t lrs_sz = strlen(ls) + 1 + strlen(rs); /* ls " " rs */
+	char *lrs = alloca(lrs_sz + 1);
+	size_t n = lg_strlcpy(lrs, ls, lrs_sz);
+	if ((ls[0] != '\0') && (rs[0] != '\0'))
+	    n += lg_strlcpy(lrs + n, " ", lrs_sz);
+	lg_strlcpy(lrs + n, rs, lrs_sz);
+	lrs[lrs_sz] = '\0';
+
+	free(ls);
+	free(rs);
+
+	return connector_list_to_expression(lrs);
+}
+
+/**
+ * Return the Category_cost array (NULL terminated) of the given disjunct.
+ * It shouldn't be freed by the caller.
+ */
+const Category_cost * disjunct_categories(Disjunct *d)
+{
+	if (d->is_category == 0) return NULL;
+	return d->category;
+}
+
+/**
+ * Return a NULL terminated array of pointers to disjuncts which are
+ * unused in the current sentence-generation linkage.
+ * Note: Only wild-card words are considered (fixed words are currently
+ * ignored).
+ * The caller has to free the returned value.
+ */
+Disjunct ** sentence_unused_disjuncts(Sentence sent)
+{
+	if ((sent == NULL) || (sent->disjunct_used == NULL)) return NULL;
+
+	unsigned int n = 0;
+	for (unsigned int i = 0; i < sent->wildcard_word_num_disjuncts; i++)
+	{
+		if (!sent->disjunct_used[i]) n++;
+	}
+	const size_t unused_d_sz = sizeof(Disjunct *) * (n + 1); /* 1 for NULL */
+	Disjunct **unused_d = malloc(unused_d_sz);
+
+	n = 0;
+	for (unsigned int i = 0; i < sent->wildcard_word_num_disjuncts; i++)
+	{
+		if (!sent->disjunct_used[i])
+			unused_d[n++] = &((Disjunct *)sent->wildcard_word_dc_memblock)[i];
+	}
+	unused_d[n] = NULL;
+
+	return unused_d;
+}
+
 /* Disjunct utilities ... */
 
 #define D_DISJ 5                        /* Verbosity level for this file. */
@@ -45,10 +128,41 @@ void free_disjuncts(Disjunct *c)
 	}
 }
 
-void free_sentence_disjuncts(Sentence sent)
+void free_categories_from_disjunct_array(Disjunct *dbase,
+                                         unsigned int num_disjuncts)
+{
+	for (Disjunct *d = dbase; d < &dbase[num_disjuncts]; d++)
+	{
+		if (d->is_category != 0)
+			free(d->category);
+	}
+}
+
+void free_categories(Sentence sent)
 {
 	if (NULL != sent->dc_memblock)
 	{
+		free_categories_from_disjunct_array(sent->dc_memblock,
+		                                    sent->num_disjuncts);
+	}
+	else
+	{
+		for (WordIdx w = 0; w < sent->length; w++)
+		{
+			for (Disjunct *d = sent->word[w].d; d != NULL; d = d->next)
+			{
+				if (d->is_category != 0)
+					free(d->category);
+			}
+		}
+	}
+}
+
+void free_sentence_disjuncts(Sentence sent, bool category_too)
+{
+	if (NULL != sent->dc_memblock)
+	{
+		if (category_too) free_categories(sent);
 		free(sent->dc_memblock);
 		sent->dc_memblock = NULL;
 	}
@@ -118,7 +232,8 @@ struct disjunct_dup_table_s
  * This is the old version that doesn't check for domination, just
  * equality.
  */
-static inline unsigned int old_hash_disjunct(disjunct_dup_table *dt, Disjunct * d)
+static inline unsigned int old_hash_disjunct(disjunct_dup_table *dt,
+                                             Disjunct * d, bool string_too)
 {
 	unsigned int i;
 	i = 0;
@@ -128,9 +243,8 @@ static inline unsigned int old_hash_disjunct(disjunct_dup_table *dt, Disjunct *
 	for (Connector *e = d->right; e != NULL; e = e->next) {
 		i = (41 * (i + e->desc->uc_num)) + (unsigned int)e->desc->lc_letters + 7;
 	}
-#if 0 /* Redundant - the connector hashing has enough entropy. */
-	i += string_hash(d->word_string);
-#endif
+	if (string_too)
+		i += string_hash(d->word_string);
 	i += (i>>10);
 
 	d->dup_hash = i;
@@ -146,7 +260,7 @@ static bool connectors_equal_prune(Connector *c1, Connector *c2)
 }
 
 /** returns TRUE if the disjuncts are exactly the same */
-static bool disjuncts_equal(Disjunct * d1, Disjunct * d2)
+static bool disjuncts_equal(Disjunct * d1, Disjunct * d2, bool ignore_string)
 {
 	Connector *e1, *e2;
 
@@ -168,6 +282,8 @@ static bool disjuncts_equal(Disjunct * d1, Disjunct * d2)
 	}
 	if ((e1 != NULL) || (e2 != NULL)) return false;
 
+	if (ignore_string) return true;
+
 	/* Save CPU time by comparing this last, since this will
 	 * almost always be true. Rarely, the strings are not from
 	 * the same string_set and hence the 2-step comparison. */
@@ -294,7 +410,7 @@ static gword_set *gword_set_union(gword_set *kept, gword_set *eliminated)
  * Takes the list of disjuncts pointed to by d, eliminates all
  * duplicates, and returns a pointer to a new list.
  */
-Disjunct *eliminate_duplicate_disjuncts(Disjunct *dw)
+Disjunct *eliminate_duplicate_disjuncts(Disjunct *dw, bool multi_string)
 {
 	unsigned int count = 0;
 	disjunct_dup_table *dt;
@@ -309,24 +425,47 @@ Disjunct *eliminate_duplicate_disjuncts(Disjunct *dw)
 	for (Disjunct *d = dw; d != NULL; d = d->next)
 	{
 		Disjunct *dx;
-		unsigned int h = old_hash_disjunct(dt, d);
+		unsigned int h = old_hash_disjunct(dt, d, /*string_too*/!multi_string);
 
 		for (dx = dt->dup_table[h]; dx != NULL; dx = dx->dup_table_next)
 		{
 			if (d->dup_hash != dx->dup_hash) continue;
-			if (disjuncts_equal(dx, d)) break;
+			if (disjuncts_equal(dx, d, multi_string)) break;
 		}
 
 		if (dx != NULL)
 		{
 			/* Discard the current disjunct. */
-			if (d->cost < dx->cost) dx->cost = d->cost;
 
-			dx->originating_gword =
-				gword_set_union(dx->originating_gword, d->originating_gword);
+			if (multi_string)
+			{
+				if (dx->num_categories == dx->num_categories_alloced - 1)
+				{
+					dx->num_categories_alloced *= 2;
+					dx->category = realloc(dx->category,
+					   sizeof(dx->category) * dx->num_categories_alloced);
+				}
+				dassert((d->category[0].num > 0) && (d->category[0].num < 64*1024),
+				        "Insane category %u", d->category[0].num);
+				dx->category[dx->num_categories].num = d->category[0].num;
+				dx->category[dx->num_categories].cost = d->cost;
+				dx->num_categories++;
+				dx->category[dx->num_categories].num = 0; /* API array terminator.*/
+			}
+			else
+			{
+				if (d->cost < dx->cost) dx->cost = d->cost;
+				dx->originating_gword =
+					gword_set_union(dx->originating_gword, d->originating_gword);
+			}
 
 			count++;
 			prev->next = d->next;
+			if (d->is_category != 0)
+			{
+				free(d->category);
+				d->is_category = 0; /* Save free() call on sentence delete. */
+			}
 		}
 		else
 		{
@@ -336,7 +475,9 @@ Disjunct *eliminate_duplicate_disjuncts(Disjunct *dw)
 		}
 	}
 
-	lgdebug(+D_DISJ+(0==count)*1000, "Killed %u duplicates\n", count);
+	lgdebug(+D_DISJ+(0==count)*1000, "w%zu: Killed %u duplicates%s\n",
+	        dw->originating_gword->o_gword->sent_wordidx, count,
+	        multi_string ? " (different word-strings)" : "");
 
 	disjunct_dup_table_delete(dt);
 	return dw;
@@ -363,7 +504,7 @@ static void prt_con(Connector *c, dyn_str * p, char dir)
 	}
 }
 
-char * print_one_disjunct(Disjunct *dj)
+char *print_one_disjunct_str(const Disjunct *dj)
 {
 	dyn_str *p = dyn_str_new();
 
@@ -508,7 +649,7 @@ void count_disjuncts_and_connectors(Sentence sent, unsigned int *dca,
  * counts because some alternatives-connectivity checks (to the middle
  * disjunct) are done in the fast-matcher. These restrictions are
  * implemented by using a different tracon ID per Gword (FIXME - this is
- * more strict then needed - a different tracon IDs per alternative would
+ * more strict then needed - a different tracon ID per alternative would
  * suffice).
  * The tracon memory sharing is currently not directly used in the
  * parsing algo besides reducing the needed CPU cache by a large factor.
@@ -602,14 +743,14 @@ static Connector *pack_connectors(Tracon_sharing *ts, Connector *origc, int dir,
 					{
 						/* This is a rare case in which a shallow and deep
 						 * connectors don't have the same nearest_word, because
-						 * a shallow connector may mach a deep connector
+						 * a shallow connector may match a deep connector
 						 * earlier. Because the nearest word is different, we
 						 * cannot share it. (Such shallow and deep tracons could
 						 * be shared separately, but because this is a rare
 						 * event there is no benefit to do that.)
 						 * Note:
 						 * In case the parsing ever depends on other Connector
-						 * fields, their will be a need to add a check for them
+						 * fields, there will be a need to add a check for them
 						 * here.
 						 * Update: farthest_word added. */
 						newc = NULL; /* Don't share it. */
@@ -626,14 +767,17 @@ static Connector *pack_connectors(Tracon_sharing *ts, Connector *origc, int dir,
 
 			if (ts->is_pruning)
 			{
-				/* Initialize for the pruning step when no sharing is done yet. */
+				/* Tracon seen for first time - initialize for the pruning stage. */
 				newc->refcount = 1;  /* No sharing yet. */
-				if (NULL != tl)
-					tl->num_cnctrs_per_word[dir][w]++;
+				if (ts->uc_seen[dir][connector_uc_num(newc)] != w)
+				{
+					ts->uc_seen[dir][connector_uc_num(newc)] = w;
+					ts->num_cnctrs_per_word[dir][w]++;
+				}
 			}
 			else
 			{
-				/* For the parsing step we need a unique ID. */
+				/* For the parsing stage we need a unique ID. */
 				newc->tracon_id = ts->next_id[dir]++;
 			}
 		}
@@ -660,10 +804,6 @@ static Connector *pack_connectors(Tracon_sharing *ts, Connector *origc, int dir,
 	return head.next;
 }
 
-/**
- * Pack the given disjunct chain in a contiguous memory block.
- * If the disjunct is NULL, return NULL.
- */
 static Disjunct *pack_disjunct(Tracon_sharing *ts, Disjunct *d, int w)
 {
 	Disjunct *newd;
@@ -672,7 +812,9 @@ static Disjunct *pack_disjunct(Tracon_sharing *ts, Disjunct *d, int w)
 	newd = (ts->dblock)++;
 	newd->word_string = d->word_string;
 	newd->cost = d->cost;
+	newd->is_category = d->is_category;
 	newd->originating_gword = d->originating_gword;
+	newd->ordinal = d->ordinal;
 
 	if (NULL == ts->tracon_list)
 		 token = (uintptr_t)d->originating_gword;
@@ -690,6 +832,10 @@ static Disjunct *pack_disjunct(Tracon_sharing *ts, Disjunct *d, int w)
 	return newd;
 }
 
+/**
+ * Pack the given disjunct chain in a contiguous memory block.
+ * If the disjunct is NULL, return NULL.
+ */
 static Disjunct *pack_disjuncts(Sentence sent, Tracon_sharing *ts,
                                 Disjunct *origd, int w)
 {
@@ -780,6 +926,25 @@ static Tracon_sharing *pack_sentence_init(Sentence sent, bool is_pruning)
 	ts->next_id[0] = ts->next_id[1] = ts->word_offset;
 	ts->last_token = (uintptr_t)-1;
 
+	if (is_pruning)
+	{
+		/* Allocate and initialize memory for finding the number of
+		 * different uppercase connector parts per direction / word, for
+		 * sizing the pruning power table. */
+		unsigned int **ncu = ts->num_cnctrs_per_word;
+		ncu[0] = malloc(2 * sent->length * sizeof(**ncu));
+		ncu[1] = ncu[0] + sent->length;
+		memset(ncu[0], 0, 2 * sent->length * sizeof(**ncu));
+
+		size_t uc_num = sent->dict->contable.num_uc;
+		ts->uc_seen[0] = malloc(2 * uc_num * sizeof(**ts->uc_seen));
+		ts->uc_seen[1] = ts->uc_seen[0] + uc_num;
+		/* Initialize w/an invalid word number in a hopefully (**uc_seen)
+		 * size independent manner.
+		 * Note that (unsigned char)-1 is currently MAX_SENTENCE+1. */
+		memset(ts->uc_seen[0], -1, 2 * uc_num * sizeof(**ts->uc_seen));
+	}
+
 	/* Encode connectors only for long-enough sentences. */
 	if (sent->length >= sent->min_len_encoding)
 	{
@@ -788,16 +953,15 @@ static Tracon_sharing *pack_sentence_init(Sentence sent, bool is_pruning)
 
 		if (is_pruning)
 		{
-			ts->tracon_list = malloc(sizeof(Tracon_list));
-			memset(ts->tracon_list, 0, sizeof(Tracon_list));
-			unsigned int **ncpw = ts->tracon_list->num_cnctrs_per_word;
+			Tracon_list *tl;
+
+			tl = ts->tracon_list = malloc(sizeof(Tracon_list));
+			memset(tl, 0, sizeof(Tracon_list));
 			for (int dir = 0; dir < 2; dir++)
 			{
-				ncpw[dir] = malloc(sent->length * sizeof(**ncpw));
-				memset(ncpw[dir], 0, sent->length * sizeof(**ncpw));
 
 				tracon_set_shallow(true, ts->csid[dir]);
-				tlsz_check(ts->tracon_list, TLSZ, dir); /* Allocate table. */
+				tlsz_check(tl, TLSZ, dir); /* Allocate table. */
 			}
 		}
 	}
@@ -821,17 +985,19 @@ void free_tracon_sharing(Tracon_sharing *ts)
 	for (int dir = 0; dir < 2; dir++)
 	{
 		if (NULL != ts->tracon_list)
-		{
-			free(ts->tracon_list->num_cnctrs_per_word[dir]);
 			free(ts->tracon_list->table[dir]);
-		}
+
 		if (NULL != ts->csid[dir])
 		{
 			tracon_set_delete(ts->csid[dir]);
 			ts->csid[dir] = NULL;
 		}
+
 	}
 
+	free(ts->uc_seen[0]);
+	free(ts->num_cnctrs_per_word[0]);
+
 	if (NULL != ts->d) free(ts->d);
 	free(ts->tracon_list);
 	ts->tracon_list = NULL;
@@ -840,7 +1006,7 @@ void free_tracon_sharing(Tracon_sharing *ts)
 }
 
 /**
- * Pack all disjunct and connectors into a one big memory block, share
+ * Pack all disjunct and connectors into one big memory block, share
  * tracon memory and generate tracon IDs (for parsing) or tracon lists
  * with reference count (for pruning). Aka "connector encoding".
  *
