@@ -16,6 +16,7 @@
 
 #include "api-structures.h"
 #include "connectors.h"
+#include "dict-common/dict-api.h"       // linkage_get_categories
 #include "dict-common/dict-affix.h"     // INFIX_MARK
 #include "dict-common/dict-defines.h"   // SUBSCRIPT_MARK
 #include "dict-common/idiom.h"
@@ -233,7 +234,8 @@ void remove_empty_words(Linkage lkg)
 		 * wg_path doesn't include them, and hence *wgp is NULL then. */
 		if ((NULL == *wgp) || ((*wgp)->sent_wordidx != i))
 		{
-			assert((NULL == cdj[i]) && lkg->sent->word[i].optional);
+			assert((NULL == cdj[i]) && lkg->sent->word[i].optional,
+			   "A matching disjunct found for a skipped optional word");
 			remap[i] = -1;
 			continue;
 		}
@@ -265,7 +267,6 @@ void remove_empty_words(Linkage lkg)
  *   computed by extract_links().  It represents the chosen disjuncts
  *   for the current linkage. It is used here to compute the chosen_words[].
  *
- *
  * wg_path[]
  *    A pointer to a NULL-terminated array of pointers to Wordgraph words.
  *    It corresponds 1-1 to the chosen_disjuncts array in Linkage structure.
@@ -307,7 +308,8 @@ void remove_empty_words(Linkage lkg)
  * Note: For historical reasons there is some overlap between the results.
  */
 #define D_CCW 8
-void compute_chosen_words(Sentence sent, Linkage linkage, Parse_Options opts)
+static void compute_chosen_words(Sentence sent, Linkage linkage,
+                                 Parse_Options opts)
 {
 	WordIdx i;   /* index of chosen_words */
 	WordIdx j;
@@ -689,7 +691,7 @@ void compute_chosen_words(Sentence sent, Linkage linkage, Parse_Options opts)
 	 * to facilitate using diff on sentence batch runs. */
 	if (test_enabled("removeZZZ"))
 	{
-		for (i=0, j=0; i<linkage->num_links; i++)
+		for (i=0; i<linkage->num_links; i++)
 		{
 			Link *lnk = &(linkage->link_array[i]);
 
@@ -739,6 +741,39 @@ void compute_chosen_words(Sentence sent, Linkage linkage, Parse_Options opts)
 }
 #undef D_CCW
 
+#define D_CGW 5
+void compute_generated_words(Sentence sent, Linkage linkage)
+{
+	Disjunct **cdjp = linkage->chosen_disjuncts;
+
+	linkage->word = malloc(linkage->num_words * sizeof(char *));
+
+	for (WordIdx i = 0; i < linkage->num_words; i++)
+	{
+		assert(cdjp[i] != NULL, "NULL disjunct in generated sentence");
+		const char *word;
+
+		Disjunct *cdj = cdjp[i];
+
+		if (cdj->is_category == 0)
+		{
+			word = cdj->word_string;
+		}
+		else
+		{
+			assert(cdj->num_categories > 0, "0 categories in disjunct");
+			word = linkage_get_disjunct_str(linkage, i);
+			size_t len = strlen(word) + sizeof("<>");
+			char *disjunct_string = alloca(len);
+			snprintf(disjunct_string, len, "<%s>", word);
+			word = string_set_add(disjunct_string, sent->string_set);
+		}
+
+		linkage->word[i] = word;
+	}
+}
+#undef D_CGW
+
 Linkage linkage_create(LinkageIdx k, Sentence sent, Parse_Options opts)
 {
 	Linkage linkage;
@@ -756,7 +791,8 @@ Linkage linkage_create(LinkageIdx k, Sentence sent, Parse_Options opts)
 	}
 
 	/* Perform remaining initialization we haven't done yet...*/
-	compute_chosen_words(sent, linkage, opts);
+	if (!IS_GENERATION(sent->dict))
+		compute_chosen_words(sent, linkage, opts);
 
 	linkage->is_sent_long = (linkage->num_words >= opts->twopass_length);
 
@@ -862,19 +898,6 @@ double linkage_get_disjunct_cost(const Linkage linkage, WordIdx w)
 	return 0.0;
 }
 
-double linkage_get_disjunct_corpus_score(const Linkage linkage, WordIdx w)
-{
-	Disjunct *dj;
-
-	if (linkage->num_words <= w) return 99.999; /* bounds-check */
-	dj = linkage->chosen_disjuncts[w];
-
-	/* dj may be null, if the word didn't participate in the parse. */
-	if (NULL == dj) return 99.999;
-
-	return lg_corpus_disjunct_score(linkage, w);
-}
-
 const char * linkage_get_word(const Linkage linkage, WordIdx w)
 {
 	if (!linkage) return NULL;
@@ -903,13 +926,6 @@ int linkage_link_cost(const Linkage linkage)
 	return linkage->lifo.link_cost;
 }
 
-double linkage_corpus_cost(const Linkage linkage)
-{
-	/* The sat solver (currently) fails to fill in info */
-	if (!linkage) return 0.0;
-	return linkage->lifo.corpus_cost;
-}
-
 /* =========== Get word sentence positions ============================== */
 
 WordIdx linkage_get_word_byte_start(const Linkage linkage, WordIdx w)
@@ -943,3 +959,13 @@ WordIdx linkage_get_word_char_end(const Linkage linkage, WordIdx w)
 	char *sentchunk = strndupa(linkage->sent->orig_sentence, pos);
 	return utf8_strlen(sentchunk);
 }
+
+const Category_cost *linkage_get_categories(const Linkage linkage, WordIdx w)
+{
+	if (NULL == linkage) return NULL;
+	if (linkage->num_words <= w) return NULL; /* bounds-check */
+
+	Disjunct *dj = linkage->chosen_disjuncts[w];
+	if (dj->is_category == 0) return NULL;
+	return dj->category;
+}
