@@ -21,11 +21,6 @@
 
 // TODO provide gword access methods!
 
-/* conditional compiling flags */
-#define INFIX_NOTATION
-    /* If defined, then we're using infix notation for the dictionary */
-    /* otherwise we're using prefix notation */
-
 /* An ordered set of gword pointers, used to indicate the source gword
  * (Wordgraph word) of disjuncts and connectors. Usually it contains only
  * one element.  However, when a duplicate disjunct is eliminated (see
@@ -117,6 +112,7 @@ typedef enum
 } Guess_mark;
 
 #define MAX_SPLITS 10   /* See split_counter below */
+#define WILDCARD_WORD "\\*" /* All the dict words (generation mode) */
 
 struct Gword_struct
 {
@@ -188,5 +184,4 @@ struct Wordgraph_pathpos_s
 	/* Only for sane_morphism(). */
 	const Gword **path; /* Linkage candidate wordgraph path */
 };
-
 #endif
