@@ -34,33 +34,23 @@
  ****************************************************************************/
 
 #include <errno.h>
+#include <limits.h>                     // CHAR_BIT
 #include <locale.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
 
-/* Used for terminal resizing */
+/* For dup2*(). */
 #ifndef _WIN32
-#include <termios.h>
-#include <sys/ioctl.h>
-#include <fcntl.h>
 #include <unistd.h>
 #else
-#include <windows.h>
-#include <wchar.h>
 #include <io.h>
-#endif /* _WIN32 */
+#endif
 
 #ifdef _MSC_VER
 #define LINK_GRAMMAR_DLL_EXPORT 0
 #endif /* _MSC_VER */
 
-#ifndef _WIN32
-#define LAST_RESORT_LOCALE "en_US.UTF-8" /* Supposing POSIX systems */
-#else
-#define LAST_RESORT_LOCALE ""            /* Use user default locale */
-#endif /* _WIN32 */
-
 #include "parser-utilities.h"
 #include "command-line.h"
 #include "lg_readline.h"
@@ -97,9 +87,9 @@ static char * get_terminal_line(char *input_string, FILE *in, FILE *out)
 	if (!running_under_cygwin)
 		pline = get_console_line();
 	else
-		pline = fgets(input_string, MAX_INPUT, in);
+		pline = fgets(input_string, MAX_INPUT_LINE, in);
 #else
-	pline = fgets(input_string, MAX_INPUT, in);
+	pline = fgets(input_string, MAX_INPUT_LINE, in);
 #endif /* _WIN32 */
 #endif /* HAVE_EDITLINE */
 
@@ -109,7 +99,7 @@ static char * get_terminal_line(char *input_string, FILE *in, FILE *out)
 static char * fget_input_string(FILE *in, FILE *out, bool check_return)
 {
 	static char *pline;
-	static char input_string[MAX_INPUT];
+	static char input_string[MAX_INPUT_LINE];
 	static bool input_pending = false;
 
 	if (input_pending)
@@ -118,12 +108,12 @@ static char * fget_input_string(FILE *in, FILE *out, bool check_return)
 		return pline;
 	}
 
-	input_string[MAX_INPUT-2] = '\0';
+	input_string[MAX_INPUT_LINE-2] = '\0';
 
 	if (((in != stdin) && !check_return) || !isatty_stdin)
 	{
 		/* Get input from a file. */
-		pline = fgets(input_string, MAX_INPUT, in);
+		pline = fgets(input_string, MAX_INPUT_LINE, in);
 	}
 	else
 	{
@@ -133,10 +123,10 @@ static char * fget_input_string(FILE *in, FILE *out, bool check_return)
 
 	if (NULL == pline) return NULL;      /* EOF or error */
 
-	if (('\0' != input_string[MAX_INPUT-2]) &&
-	    ('\n' != input_string[MAX_INPUT-2]))
+	if (('\0' != input_string[MAX_INPUT_LINE-2]) &&
+	    ('\n' != input_string[MAX_INPUT_LINE-2]))
 	{
-		prt_error("Warning: Input line too long (>%d)\n", MAX_INPUT-1);
+		prt_error("Warning: Input line too long (>%d)\n", MAX_INPUT_LINE-1);
 		/* TODO: Ignore it and its continuation part(s). */
 	}
 
@@ -175,6 +165,9 @@ static void process_linkage(Linkage linkage, Command_Options* copts)
 	}
 	if (copts->display_on)
 	{
+#ifdef SIGWINCH
+		set_screen_width(copts);
+#endif
 		string = linkage_print_diagram(linkage, copts->display_walls, copts->screen_width);
 		fprintf(stdout, "%s", string);
 		linkage_free_diagram(string);
@@ -200,12 +193,6 @@ static void process_linkage(Linkage linkage, Command_Options* copts)
 		fprintf(stdout, "%s", string);
 		linkage_free_links_and_domains(string);
 	}
-	if (copts->display_senses)
-	{
-		string = linkage_print_senses(linkage);
-		fprintf(stdout, "%s", string);
-		linkage_free_senses(string);
-	}
 	if (copts->display_disjuncts)
 	{
 		string = linkage_print_disjuncts(linkage);
@@ -318,7 +305,6 @@ static const char *process_some_linkages(FILE *in, Sentence sent,
 {
 	int i, num_to_query, num_to_display, num_displayed;
 	Linkage linkage;
-	double corpus_cost;
 	Parse_Options opts = copts->popts;
 	int display_max = DISPLAY_MAX;
 	bool auto_next_linkage = false;
@@ -390,22 +376,10 @@ static const char *process_some_linkages(FILE *in, Sentence sent,
 				fprintf(stdout, "\tLinkage %d, ", num_displayed+1);
 			}
 
-			corpus_cost = linkage_corpus_cost(linkage);
-			if (corpus_cost < 0.0f)
-			{
-				fprintf(stdout, "cost vector = (UNUSED=%d DIS=%5.2f LEN=%d)\n",
-				       linkage_unused_word_cost(linkage),
-				       linkage_disjunct_cost(linkage),
-				       linkage_link_cost(linkage));
-			}
-			else
-			{
-				fprintf(stdout, "cost vector = (CORP=%6.4f UNUSED=%d DIS=%5.2f LEN=%d)\n",
-				       corpus_cost,
-				       linkage_unused_word_cost(linkage),
-				       linkage_disjunct_cost(linkage),
-				       linkage_link_cost(linkage));
-			}
+			fprintf(stdout, "cost vector = (UNUSED=%d DIS=%5.2f LEN=%d)\n",
+			        linkage_unused_word_cost(linkage),
+			        linkage_disjunct_cost(linkage),
+			        linkage_link_cost(linkage));
 		}
 
 		process_linkage(linkage, copts);
@@ -490,7 +464,7 @@ static void batch_process_some_linkages(Label label,
  * If input_string is !command, try to issue it.
  */
 
-static char special_command(char *input_string, Command_Options* copts, Dictionary dict)
+static int special_command(char *input_string, Command_Options* copts, Dictionary dict)
 {
 	if (input_string[0] == COMMENT_CHAR) return 'c';
 	if (input_string[0] == '!')
@@ -514,19 +488,6 @@ static Label strip_off_label(char * input_string)
 	}
 }
 
-static void setup_panic_parse_options(Parse_Options opts)
-{
-	parse_options_set_disjunct_cost(opts, 4.0f);
-	parse_options_set_min_null_count(opts, 1);
-	parse_options_set_max_null_count(opts, 100);
-	parse_options_set_max_parse_time(opts, 60);
-	parse_options_set_islands_ok(opts, false);
-	parse_options_set_short_length(opts, 12);
-	parse_options_set_all_short_connectors(opts, true);
-	parse_options_set_linkage_limit(opts, 100);
-	parse_options_set_spell_guess(opts, 0);
-}
-
 static int divert_stdio(FILE *from, FILE *to)
 {
 	const int origfd = dup(fileno(from));
@@ -579,58 +540,6 @@ static void print_usage(FILE *out, char *argv0, Command_Options *copts, int exit
 	exit(exit_value);
 }
 
-/**
- * On Unix, this checks for the current window size,
- * and sets the output screen width accordingly.
- */
-static void check_winsize(Command_Options* copts)
-{
-	if (!isatty_stdout) return;
-	int fd = fileno(stdout);
-#ifdef _WIN32
-	HANDLE console;
-	CONSOLE_SCREEN_BUFFER_INFO info;
-
-	/* Create a handle to the console screen. */
-	console = (HANDLE)_get_osfhandle(fd);
-	if (!console || (console == INVALID_HANDLE_VALUE)) goto fail;
-
-	/* Calculate the size of the console window. */
-	if (GetConsoleScreenBufferInfo(console, &info) == 0) goto fail;
-
-	copts->screen_width = (size_t)(info.srWindow.Right - info.srWindow.Left + 1);
-	return;
-
-fail:
-	copts->screen_width = 80;
-	return;
-#else
-	struct winsize ws;
-
-	/* If there is no controlling terminal, the fileno will fail. This
-	 * seems to happen while building docker images, I don't know why.
-	 */
-	if (fd < 0) return;
-
-	if (0 != ioctl(fd, TIOCGWINSZ, &ws))
-	{
-		perror("stdout: ioctl TIOCGWINSZ");
-		return;
-	}
-
-	/* printf("rows %i\n", ws.ws_row); */
-	/* printf("cols %i\n", ws.ws_col); */
-
-	/* Set the screen width only if the returned value seems
-	 * rational: it's positive and not insanely tiny.
-	 */
-	if ((10 < ws.ws_col) && (16123 > ws.ws_col))
-	{
-		copts->screen_width = ws.ws_col;
-	}
-#endif /* _WIN32 */
-}
-
 int main(int argc, char * argv[])
 {
 	FILE            *input_fh = stdin;
@@ -668,42 +577,20 @@ int main(int argc, char * argv[])
 	win32_set_utf8_output();
 #endif /* _WIN32 */
 
-#if LATER
-	/* Try to catch the SIGWINCH ... except this is not working. */
-	struct sigaction winch_act;
-	winch_act.sa_handler = winch_handler;
-	winch_act.sa_sigaction = NULL;
-	sigemptyset (&winch_act.sa_mask);
-	winch_act.sa_flags = 0;
-	sigaction (SIGWINCH, &winch_act, NULL);
-#endif
-
-	copts = command_options_create();
-	if (copts == NULL || copts->panic_opts == NULL)
-	{
-		prt_error("Fatal error: unable to create parse options\n");
-		exit(-1);
-	}
-	opts = copts->popts;
-
-	setup_panic_parse_options(copts->panic_opts);
-	copts->panic_mode = true;
-
-	parse_options_set_max_parse_time(opts, 30);
-	parse_options_set_linkage_limit(opts, 1000);
-	parse_options_set_min_null_count(opts, 0);
-	parse_options_set_max_null_count(opts, 0);
-	parse_options_set_short_length(opts, 16);
-	parse_options_set_islands_ok(opts, false);
-	parse_options_set_display_morphology(opts, false);
-
-	save_default_opts(copts); /* Options so far are the defaults */
-
 	if ((argc > 1) && (argv[1][0] != '-')) {
 		/* The dictionary is the first argument if it doesn't begin with "-" */
 		language = argv[1];
 	}
 
+	copts = command_options_create();
+
+	/* First set the debug options, to allow dictionary-related debug. */
+	const char * const debug_vars[] = { "verbosity", "debug", "test" };
+	const int debug_vars_min_len[] = { 1, 2, 2 }; /* Ambiguous if shorter */
+	unsigned long long argv_done = 0;             /* Process them only once */
+
+	/* Process options used by GNU programs. */
+	int quiet_start = 0; /* Iff > 0, inhibit the initial messages */
 	for (int i = 1; i < argc; i++)
 	{
 		if (strcmp("--help", argv[i]) == 0)
@@ -717,23 +604,45 @@ int main(int argc, char * argv[])
 			printf("%s\n", linkgrammar_get_configuration());
 			exit(0);
 		}
+
+		if ((strcmp("--quiet", argv[i]) == 0) ||
+		    (strcmp("--silent", argv[i]) == 0))
+		{
+			quiet_start = i;
+		}
 	}
 
-	/* Process command line variable-setting commands (only). */
+	/* Process debug command line variable-setting commands (only). */
 	for (int i = 1; i < argc; i++)
 	{
+		if (i == quiet_start) continue;
 		if (argv[i][0] == '-')
 		{
-			const char *var = argv[i] + ((argv[i][1] != '-') ? 1 : 2);
-			if ((var[0] != '!') && (0 > issue_special_command(var, copts, NULL)))
-				print_usage(stderr, argv[0], copts, -1);
-		}
-		else if (i != 1)
-		{
-			prt_error("Fatal error: Unknown argument '%s'.\n", argv[i]);
-			print_usage(stderr, argv[0], copts, -1);
+			char *var = strdup(argv[i] + ((argv[i][1] != '-') ? 1 : 2));
+			char *eqpos = strchr(var, '=');
+
+			if (eqpos != NULL)
+			{
+				*eqpos = '\0';
+
+				for (size_t j = 0; j < sizeof(debug_vars)/sizeof(debug_vars[0]); j++)
+				{
+					if (eqpos - var < debug_vars_min_len[j]) continue;
+					if (strncasecmp(debug_vars[j], var, eqpos - var) != 0) continue;
+					*eqpos = '=';
+					if (0 > issue_special_command(var, copts, NULL))
+						print_usage(stderr, argv[0], copts, -1);
+					if (i < (int)sizeof(argv_done) * CHAR_BIT)
+						argv_done |= (1LL<<(i-1));
+					else
+						prt_error("Warning: Too many arguments (%d).\n", i);
+					break;
+				}
+			}
+			free(var);
 		}
 	}
+	/* End of debug options setup. */
 
 	if (language && *language)
 	{
@@ -754,6 +663,86 @@ int main(int argc, char * argv[])
 		}
 	}
 
+	if (copts == NULL || copts->popts == NULL)
+	{
+		prt_error("Fatal error: unable to create parse options\n");
+		exit(-1);
+	}
+	opts = copts->popts;
+
+	parse_options_set_max_parse_time(opts, 30);
+	parse_options_set_linkage_limit(opts, 1000);
+	parse_options_set_min_null_count(opts, 0);
+	parse_options_set_max_null_count(opts, 0);
+	parse_options_set_short_length(opts, 16);
+	parse_options_set_islands_ok(opts, false);
+	parse_options_set_display_morphology(opts, false);
+
+	/* Get the panic disjunct cost from the dictionary. */
+	const char *panic_max_cost_str =
+		linkgrammar_get_dict_define(dict, LG_PANIC_DISJUNCT_COST);
+	if (panic_max_cost_str != NULL)
+	{
+		const char *locale =  setlocale(LC_NUMERIC, "C");
+		char *err;
+		double panic_max_cost = strtod(panic_max_cost_str, &err);
+		setlocale(LC_NUMERIC, locale);
+
+		if ('\0' == *err)
+		{
+			copts->panic.max_cost = panic_max_cost;
+		}
+		else
+		{
+			prt_error("Warning: Unparsable "LG_PANIC_DISJUNCT_COST " \"%s\" "
+			          "in the dictionary\n", panic_max_cost_str);
+		}
+	}
+
+	/* This is not absolutely needed because the library set it after
+	 * the first parse. However, it is initialized here so !variables,
+	 * "link-parser DICT -variables" and "!help cost-max" show the
+	 * correct value. */
+	parse_options_set_disjunct_cost(opts,
+	   linkgrammar_get_dict_max_disjunct_cost(dict));
+
+	/* Remember the debug setting, because we temporary neglect it below. */
+	int verbosity_tmp = parse_options_get_verbosity(opts);
+	char *debug_tmp = strdup(parse_options_get_debug(opts));
+	char *test_tmp = strdup(parse_options_get_test(opts));
+
+	parse_options_set_verbosity(opts, 1); /* XXX assuming 1 is the default */
+	parse_options_set_debug(opts, "");
+	parse_options_set_test(opts, "");
+	save_default_opts(copts); /* Options so far are the defaults */
+
+	/* Restore the debug setting. */
+	parse_options_set_verbosity(opts, verbosity_tmp);
+	parse_options_set_debug(opts, debug_tmp);
+	parse_options_set_test(opts, test_tmp);
+	free(debug_tmp);
+	free(test_tmp);
+
+	/* Process non-debug command line variable-setting commands (only). */
+	for (int i = 1; i < argc; i++)
+	{
+		if (i == quiet_start) continue;
+		if ((i < (int)sizeof(argv_done) * CHAR_BIT) &&
+		    (argv_done & (1LL<<(i-1)))) continue;
+
+		if (argv[i][0] == '-')
+		{
+			const char *var = argv[i] + ((argv[i][1] != '-') ? 1 : 2);
+			if ((var[0] != '!') && (0 > issue_special_command(var, copts, NULL)))
+				print_usage(stderr, argv[0], copts, -1);
+		}
+		else if (i != 1)
+		{
+			prt_error("Fatal error: Unknown argument '%s'.\n", argv[i]);
+			print_usage(stderr, argv[0], copts, -1);
+		}
+	}
+
 	/* Process the command line '!' commands */
 	for (int i = 1; i < argc; i++)
 	{
@@ -764,13 +753,16 @@ int main(int argc, char * argv[])
 		}
 	}
 
-	check_winsize(copts);
+	initialize_screen_width(copts);
 
-	prt_error("Info: Dictionary version %s, locale %s\n",
-		linkgrammar_get_dict_version(dict),
-		linkgrammar_get_dict_locale(dict));
-	prt_error("Info: Library version %s. Enter \"!help\" for help.\n",
-		linkgrammar_get_version());
+	if ((parse_options_get_verbosity(opts)) > 0 && (quiet_start == 0))
+	{
+		prt_error("Info: Dictionary version %s, locale %s\n",
+			linkgrammar_get_dict_version(dict),
+			linkgrammar_get_dict_locale(dict));
+		prt_error("Info: Library version %s. Enter \"!help\" for help.\n",
+			linkgrammar_get_version());
+	}
 
 	/* Main input loop */
 	while (true)
@@ -787,7 +779,6 @@ int main(int argc, char * argv[])
 		test = parse_options_get_test(opts);
 
 		input_string = fget_input_string(input_fh, stdout, /*check_return*/false);
-		check_winsize(copts);
 
 		if (NULL == input_string)
 		{
@@ -802,7 +793,7 @@ int main(int argc, char * argv[])
 
 		/* Discard whitespace characters from end of string. */
 		for (char *p = &input_string[strlen(input_string)-1];
-		     (p > input_string) && strchr(WHITESPACE, *p) ; p--)
+		     (p > input_string) && strchr(WHITESPACE, *p); p--)
 		{
 			*p = '\0';
 		}
@@ -811,7 +802,8 @@ int main(int argc, char * argv[])
 		if (strspn(input_string, WHITESPACE) == strlen(input_string))
 			continue;
 
-		char command = special_command(input_string, copts, dict);
+		set_screen_width(copts);
+		int command = special_command(input_string, copts, dict);
 		if ('e' == command) break;    /* It was an exit command */
 		if ('c' == command) continue; /* It was another command */
 		if (-1 == command) continue;  /* It was a bad command */
@@ -820,39 +812,42 @@ int main(int argc, char * argv[])
 		 * otherwise ... */
 		if ('f' == command)
 		{
-			char * filename = &input_string[strcspn(input_string, WHITESPACE)] + 1;
-			int fnlen = strlen(filename);
-
-			if (0 == fnlen)
+			char *command_end = &input_string[strcspn(input_string, WHITESPACE)];
+			char *filename = &command_end[strspn(command_end, WHITESPACE)];
+			if (filename[0] == '\0')
 			{
 				prt_error("Error: Missing file name argument\n");
 				continue;
 			}
 
-			if ('\n' == filename[fnlen-1]) filename[fnlen-1] = '\0';
+			char *eh_filename = expand_homedir(filename);
 
 			struct stat statbuf;
-			if ((0 == stat(filename, &statbuf)) && statbuf.st_mode & S_IFDIR)
+			if ((0 == stat(eh_filename, &statbuf)) && statbuf.st_mode & S_IFDIR)
 			{
-				prt_error("Error: Cannot open %s: %s\n",
-				        filename, strerror(EISDIR));
-				continue;
+				errno = EISDIR;
+				goto open_error;
 			}
 
-			input_fh = fopen(filename, "r");
+			input_fh = fopen(eh_filename, "r");
 
 			if (NULL == input_fh)
 			{
-				prt_error("Error: Cannot open %s: %s\n", filename, strerror(errno));
 				input_fh = stdin;
-				continue;
+				goto open_error;
 			}
+
+			free(eh_filename);
+			continue;
+
+open_error:
+			prt_error("Error: Cannot open %s: %s\n", eh_filename, strerror(errno));
+			free(eh_filename);
 			continue;
 		}
 
-
 		if (!copts->batch_mode) batch_in_progress = false;
-		if ('\0' != test[0])
+		if ('\0' != test[0] && !test_enabled(test, "@"))
 		{
 			/* In batch mode warn only once.
 			 * In auto-next-linkage mode don't warn at all. */
@@ -918,11 +913,11 @@ int main(int argc, char * argv[])
 		}
 
 		/* First parse with the default disjunct_cost as set by the library
-		 * (currently 2.7). Usually parse here with no null links.
+		 * (typically 2.7). Usually parse here with no null links.
 		 * However, if "-test=one-step-parse" is used and we are said to
 		 * parse with null links, allow parsing here with null links too. */
 		bool one_step_parse = !copts->batch_mode && copts->allow_null &&
-		                    test_enabled(test, "one-step-parse");
+		                      test_enabled(test, "one-step-parse");
 		int max_null_count = one_step_parse ? sentence_length(sent) : 0;
 
 		parse_options_set_min_null_count(opts, 0);
@@ -940,64 +935,51 @@ int main(int argc, char * argv[])
 			continue;
 		}
 
-#if 0
-		/* Try again, this time omitting the requirement for
-		 * definite articles, etc. This should allow for the parsing
-		 * of newspaper headlines and other clipped speech.
-		 *
-		 * XXX Unfortunately, this also allows for the parsing of
-		 * all sorts of ungrammatical sentences which should not
-		 * parse, and leads to bad parses of many other unparsable
-		 * but otherwise grammatical sentences.  Thus, this trick
-		 * pretty much fails; we leave it here to document the
-		 * experiment.
-		 */
-		if (num_linkages == 0)
+		if (!(copts->panic_mode && parse_options_resources_exhausted(opts)))
 		{
-			parse_options_set_disjunct_cost(opts, 4.5);
-			num_linkages = sentence_parse(sent, opts);
-			if (num_linkages < 0) continue;
-		}
+#if 0
+			/* Try again, this time omitting the requirement for
+			 * definite articles, etc. This should allow for the parsing
+			 * of newspaper headlines and other clipped speech.
+			 *
+			 * XXX Unfortunately, this also allows for the parsing of
+			 * all sorts of ungrammatical sentences which should not
+			 * parse, and leads to bad parses of many other unparsable
+			 * but otherwise grammatical sentences.  Thus, this trick
+			 * pretty much fails; we leave it here to document the
+			 * experiment.
+			 */
+			if (num_linkages == 0)
+			{
+				parse_options_set_disjunct_cost(opts, 4.5);
+				num_linkages = sentence_parse(sent, opts);
+				if (num_linkages < 0) continue;
+			}
 #endif /* 0 */
 
-		/* Try using a larger list of disjuncts */
-		/* XXX FIXME: the lg_expand_disjunct_list() routine is not
-		 * currently a part of the public API; it should be made so,
-		 * or this expansion idea should be abandoned... not sure which.
-		 */
-		if ((num_linkages == 0) && parse_options_get_use_cluster_disjuncts(opts))
-		{
-			int expanded;
-			if (verbosity > 0) fprintf(stdout, "No standard linkages, expanding disjunct set.\n");
-			parse_options_set_disjunct_cost(opts, 3.9);
-			expanded = lg_expand_disjunct_list(sent);
-			if (expanded)
+			/* If asked to show bad linkages, then show them. */
+			if ((num_linkages == 0) && (!copts->batch_mode))
 			{
-				num_linkages = sentence_parse(sent, opts);
+				if (copts->display_bad)
+				{
+					num_linkages = sentence_num_linkages_found(sent);
+				}
 			}
-			if (0 < num_linkages) printf("Got One !!!!!!!!!!!!!!!!!\n");
-		}
 
-		/* If asked to show bad linkages, then show them. */
-		if ((num_linkages == 0) && (!copts->batch_mode))
-		{
-			if (copts->display_bad)
+			/* Now parse with null links */
+			if (!one_step_parse && num_linkages == 0)
 			{
-				num_linkages = sentence_num_linkages_found(sent);
-			}
-		}
-
-		/* Now parse with null links */
-		if (!one_step_parse && num_linkages == 0 && !copts->batch_mode)
-		{
-			if (verbosity > 0) fprintf(stdout, "No complete linkages found.\n");
-
-			if (copts->allow_null)
-			{
-				/* XXX should use expanded disjunct list here too */
-				parse_options_set_min_null_count(opts, 1);
-				parse_options_set_max_null_count(opts, sentence_length(sent));
-				num_linkages = sentence_parse(sent, opts);
+				if (verbosity > 0) fprintf(stdout, "No complete linkages found.\n");
+				if (!copts->batch_mode)
+				{
+					if (copts->allow_null)
+					{
+						/* XXX should use expanded disjunct list here too */
+						parse_options_set_min_null_count(opts, 1);
+						parse_options_set_max_null_count(opts, sentence_length(sent));
+						num_linkages = sentence_parse(sent, opts);
+					}
+				}
 			}
 		}
 
@@ -1010,34 +992,30 @@ int main(int argc, char * argv[])
 				fprintf(stdout, "Memory is exhausted!\n");
 		}
 
-		if ((num_linkages == 0) &&
-			copts->panic_mode &&
-			parse_options_resources_exhausted(opts))
+		if (copts->panic_mode && parse_options_resources_exhausted(opts))
 		{
-			/* print_total_time(opts); */
 			batch_errors++;
-			if (verbosity > 0) fprintf(stdout, "Entering \"panic\" mode...\n");
-			/* If the parser used was the SAT solver, set the panic parser to
-			 * it too.
-			 * FIXME? Currently, the SAT solver code is not too useful in
-			 * panic mode since it doesn't handle parsing with null words, so
-			 * using the regular parser in that case could be beneficial.
-			 * However, this currently causes a crash due to a memory
-			 * management mess. */
-			parse_options_set_use_sat_parser(copts->panic_opts,
-				parse_options_get_use_sat_parser(opts));
-			parse_options_reset_resources(copts->panic_opts);
-			parse_options_set_verbosity(copts->panic_opts, verbosity);
-			(void)sentence_parse(sent, copts->panic_opts);
 			if (verbosity > 0)
 			{
-				if (parse_options_timer_expired(copts->panic_opts))
+				fprintf(stdout, "Entering \"panic\" mode...\n");
+			}
+
+			setup_panic_parse_options(copts, sentence_length(sent));
+			(void)sentence_parse(sent, opts);
+			if (verbosity > 0)
+			{
+				if (parse_options_timer_expired(copts->popts))
 					fprintf(stdout, "Panic timer is expired!\n");
 			}
+			put_local_vars_in_opts(copts); /* Undo setup_panic_parse_options() */
 		}
 
 		if (verbosity > 1) parse_options_print_total_time(opts);
 
+#ifndef SIGWINCH
+		set_screen_width(copts);
+#endif
+
 		const char *rc = "";
 		if (copts->batch_mode)
 		{
