@@ -11,20 +11,101 @@
 /*                                                                         */
 /***************************************************************************/
 
+/* Used for terminal resizing */
+#ifndef _WIN32
+#endif /* _WIN32 */
+
 #ifdef _WIN32
 #include <windows.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <fcntl.h>
+#include <wchar.h>
 #include <io.h>
 #include <shellapi.h>                    /* CommandLineToArgvW() */
-#ifdef __MINGW32__
-#include <malloc.h>
-#endif /* __MINGW32__ */
+#else
+#include <pwd.h>
+#include <signal.h>
+#include <sys/ioctl.h>
+#include <sys/types.h>
+#include <termios.h>
+#include <unistd.h>
+#endif /* _WIN32 */
+
 #include <errno.h>
+#include <fcntl.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 
 #include "parser-utilities.h"
 
+/**
+ * Expand an initial '~' to home directory.
+ *
+ * @param filename The filename to expand. Its length must be >= 1.
+ * @return A newly-allocated filename, expanded if possible. Freed by caller.
+ *
+ * Note: ~user is not supported here on Windows.
+ */
+char *expand_homedir(const char *filename)
+{
+	if (filename[0] != '~') return strdup(filename);
+
+#ifndef _WIN32
+	char *user = NULL;
+	const char *user_end = &filename[strcspn(filename, "/")];
+	if (user_end != &filename[1])
+	{
+		user = strdup(filename + 1);
+		user[user_end - filename - 1] = '\0';
+	}
+#endif /* _WIN32 */
+
+#ifdef _WIN32
+	char *home;
+
+	const char *homepath = getenv("HOMEPATH");
+	if ((homepath == NULL) || (homepath[0] == '\0')) return strdup(filename);
+	const char *homedrive = getenv("HOMEDRIVE");
+	if (homedrive == NULL) homedrive = "";
+
+	home = malloc(strlen(homepath) + strlen(homedrive) + 1);
+	strcpy(home, homedrive);
+	strcat(home, homepath);
+	filename++;
+#else
+	const char *home;
+
+	if (user == NULL)
+	{
+		home = getenv("HOME");
+		if ((home == NULL) || (home[0] == '\0')) return strdup(filename);
+		filename++;
+	}
+	else
+	{
+		struct passwd *pwd;
+		pwd = getpwnam(user);
+		free(user);
+		if (pwd == NULL) return strdup(filename);
+		home = pwd->pw_dir;
+		filename = user_end;
+	}
+#endif
+
+	size_t filename_len = strlen(filename);
+	size_t home_len = strlen(home);
+
+	char *eh_filename = malloc(home_len + filename_len + 1);
+	memcpy(eh_filename, home, home_len);
+	memcpy(eh_filename + home_len, filename, filename_len + 1);
+
+#ifdef _WIN32
+	free(home);
+#endif
+
+	return eh_filename;
+}
+
+#ifdef _WIN32
 /**
  * Get a line from the console in UTF-8.
  * This function bypasses the code page conversion and reads Unicode
@@ -34,9 +115,9 @@
 char *get_console_line(void)
 {
 	static HANDLE console_handle = NULL;
-	wchar_t winbuf[MAX_INPUT];
+	wchar_t winbuf[MAX_INPUT_LINE];
 	/* Worst-case: 4 bytes per UTF-8 char, 1 UTF-8 char per wchar_t char. */
-	static char utf8inbuf[MAX_INPUT*4+1];
+	static char utf8inbuf[MAX_INPUT_LINE*4+1];
 
 	if (NULL == console_handle)
 	{
@@ -50,7 +131,7 @@ char *get_console_line(void)
 	}
 
 	DWORD nchar;
-	if (!ReadConsoleW(console_handle, &winbuf, MAX_INPUT-sizeof(wchar_t), &nchar, NULL))
+	if (!ReadConsoleW(console_handle, &winbuf, MAX_INPUT_LINE-sizeof(wchar_t), &nchar, NULL))
 	{
 		prt_error("Error: ReadConsoleW: Error %d\n", (int)GetLastError());
 		return NULL;
@@ -253,6 +334,103 @@ char **argv2utf8(int argc)
 }
 #endif /* _WIN32 */
 
+static unsigned int screen_width = INITIAL_SCREEN_WIDTH;
+/**
+ * Query the system for the current tty/console width.
+ * May be called directly or as SIGWINCH handler (if SIGWINCH exists).
+ * Put the found screen width in a global variable \c screen_width;
+ * @param sig Used only for type check in case SIGWINCH exists.
+ * @return Void.
+ */
+static void get_screen_width(int sig)
+{
+	static int isatty_stdout = -1;
+
+	if (isatty_stdout == -1) isatty_stdout = isatty(fileno(stdout));
+	if (!isatty_stdout) return;
+
+	int fd = fileno(stdout);
+
+#ifdef _WIN32
+	HANDLE console;
+	CONSOLE_SCREEN_BUFFER_INFO info;
+
+	/* Create a handle to the console screen. */
+	console = (HANDLE)_get_osfhandle(fd);
+	if (!console || (console == INVALID_HANDLE_VALUE)) return;
+
+	/* Calculate the size of the console window. */
+	if (GetConsoleScreenBufferInfo(console, &info) == 0) return;
+
+	screen_width = info.dwSize.X;
+	return;
+#else
+	struct winsize ws;
+
+	/* If there is no controlling terminal, the fileno will fail. This
+	 * seems to happen while building docker images, I don't know why.
+	 */
+	if (fd < 0) return;
+
+	if (0 != ioctl(fd, TIOCGWINSZ, &ws))
+	{
+		perror("stdout: ioctl TIOCGWINSZ");
+		return;
+	}
+
+	/* Set the screen width only if the returned value seems
+	 * rational: it's positive and not insanely tiny.
+	 */
+	if ((10 < ws.ws_col) && (16123 > ws.ws_col))
+		screen_width = ws.ws_col;
+#endif /* _WIN32 */
+}
+
+/**
+ * Set the "width" user variable from the global \c screen_size.
+ * If SIGWINCH is not defined, poll the screen width beforehand.
+ */
+void set_screen_width(Command_Options* copts)
+{
+#if !defined SIGWINCH || defined _WIN32
+	get_screen_width(0);
+#endif
+
+	copts->screen_width = screen_width;
+}
+
+#ifdef INTERRUPT_EXIT
+static void interrupt_exit(int n)
+{
+	exit(128+n);
+}
+#endif
+
+void initialize_screen_width(Command_Options *copts)
+{
+#ifdef SIGWINCH
+#if HAVE_SIGACTION
+	struct sigaction winch_act = { 0 };
+	winch_act.sa_handler = get_screen_width;
+	sigemptyset(&winch_act.sa_mask);
+	winch_act.sa_flags = 0;
+	if (sigaction(SIGWINCH, &winch_act, NULL) == -1)
+		perror("sigaction SIGWINCH");
+#else
+	if (signal(SIGWINCH, get_screen_width) == SIG_ERR)
+		perror("signal SIGWINCH");
+#endif /* HAVE_SIGACTION */
+#endif /* SIGWINCH */
+
+#ifdef INTERRUPT_EXIT
+	(void)signal(SIGINT, interrupt_exit);
+	(void)signal(SIGTERM, interrupt_exit);
+#endif
+
+	get_screen_width(0);
+	set_screen_width(copts);
+}
+
 #ifdef __MINGW32__
 /*
  * A workaround for printing UTF-8 on the console.
